import numpy as np
from figaro.exceptions import FIGAROException

def MC_integral(p, q, n_draws = 1e4, error = True):
    """
    Monte Carlo integration using FIGARO reconstructions.
        ∫p(x)q(x)dx ~ ∑p(x_i)/N with x_i ~ q(x)
    
    p(x) must have a pdf() method and q(x) must have a rvs() method.
    Lists of p and q are also accepted.
    
    Arguments:
        list or class instance p: the probability density to evaluate. Must have a pdf() method.
        list or class instance q: the probability density to sample from. Must have a rvs() method.
        int n_draws:              number of MC draws
        bool error:               whether to return the uncertainty on the integral value or not.
    
    Return:
        double: integral value
        double: uncertainty (if error = True)
    """
    # Check that both p and q are iterables or callables:
    if not ((hasattr(p, 'pdf') or np.iterable(p)) and (hasattr(q, 'rvs') or np.iterable(q))):
        raise FIGAROException("p and q must be list of callables or having pdf/rvs methods")
    # Number of p draws and methods check
    iter_p = False
    iter_q = False
    if np.iterable(p):
        if not np.alltrue([hasattr(pi, 'pdf') for pi in p]):
            raise FIGAROException("p must have pdf method")
        n_p = len(p)
        np.random.shuffle(p)
        iter_p = True
    else:
        if not hasattr(p, 'pdf'):
            raise FIGAROException("p must have pdf method")
    # Number of q draws and methods check
    if np.iterable(q):
        if not np.alltrue([hasattr(qi, 'rvs') for qi in q]):
            raise FIGAROException("q must have rvs method")
        n_q = len(q)
        np.random.shuffle(q)
        iter_q = True
    else:
        if not hasattr(q, 'rvs'):
            raise FIGAROException("q must have rvs method")

    n_draws = int(n_draws)
    # Integrals
    if iter_p and iter_q:
        shortest = np.min([n_p, n_q])
        probabilities = np.array([pi.pdf(qi.rvs(n_draws)) for pi, qi in zip(p[:shortest], q[:shortest])])
    elif iter_q and not iter_p:
        probabilities = np.array([p.pdf(qi.rvs(n_draws)) for qi in q])
    elif iter_p and not iter_q:
        samples = q.rvs(n_draws)
        probabilities = np.array([pi.pdf(samples) for pi in p])
    else:
        probabilities = np.atleast_2d(p.pdf(q.rvs(n_draws)))
    means = probabilities.mean(axis = 1)
    I = means.mean()
    if not error:
        return I
    mc_error = (probabilities.var(axis = 1)/n_draws).mean()
    figaro_error = means.var()/len(means)
    return I, np.sqrt(mc_error + figaro_error)



import numpy as np
from numba import njit, prange
from numba.extending import get_cython_function_address
import ctypes

@njit
def inv_jit(M):
  return np.linalg.inv(M)

@njit
def dot_jit(v1, v2):
  return np.dot(v1, v2)

@njit
def logdet_jit(M):
    return np.log(np.linalg.det(M))

@njit
def logsumexp_jit(a):
    a_max = np.max(a)
    return np.log(np.sum(np.exp(a - a_max))) + a_max

@njit
def logsumexp_jit_weighted(a, b):
    a_max = np.max(a)
    tmp = b * np.exp(a - a_max)
    return np.log(np.sum(tmp)) + a_max

@njit
def log_add(x, y):
    if x >= y:
        return x+log1p_jit(np.exp(y-x))
    else:
        return y+log1p_jit(np.exp(x-y))

@njit
def outer_jit(x, y):
    return np.outer(x, y)
    
@njit
def divide_jit(x, y):
    return np.divide(x, y)

@njit
def diag_jit(m):
    return np.diag(m)

@njit
def rescale_matrix(S, n):
    std      = np.sqrt(diag_jit(S))
    std_filt = std
    std_filt[std == 0.] = 1. # avoids NaNs due to (0/0)*0
    rho = divide_jit(S, outer_jit(std_filt,std_filt))
    return rho * outer_jit(std/np.sqrt(n), std/np.sqrt(n))

@njit
def eigh_jit(m):
    return np.linalg.eigh(m)

@njit
def log1p_jit(x):
    return np.log1p(x)

"""
See https://stackoverflow.com/a/54855769
Wrapper (based on https://github.com/numba/numba/issues/3086) for scipy's cython implementation of gammaln.
"""

_PTR = ctypes.POINTER
_dble = ctypes.c_double
_ptr_dble = _PTR(_dble)

addr = get_cython_function_address("scipy.special.cython_special", "gammaln")
functype = ctypes.CFUNCTYPE(_dble, _dble)
gammaln_float64 = functype(addr)

@njit
def gammaln_jit(x):
    return gammaln_float64(x)



import traceback as tb
import numpy
import sys

class FIGAROException(Exception):
    pass

def except_hook(exctype, value, traceback):
    tb_last   = list(tb.walk_tb(traceback))[-1][0] # Get last call from traceback (function that raised the exception)
    try:
        tb_s2last = list(tb.walk_tb(traceback))[-2][0] # Get second-to-last call from traceback (hopefully FIGARO function, to be checked)
    except:
        tb_s2last = ""
    # Check if error is due to some known improper usage of code
    #-----------#
    # Sample outside boundaries
    if exctype == ValueError and tb_last.f_code.co_name == "numpy.random.mtrand.RandomState.choice" and tb_s2last.f_code.co_name == "_assign_to_cluster":
        sys.__excepthook__(exctype, value, traceback)
        print("\nFIGAROException: you probably have a sample that falls outside the given boundaries\n")
    elif exctype == numpy.linalg.LinAlgError and tb_last.f_code.co_name == "_check_finite_matrix" and tb_s2last.f_code.co_name == "_log_predictive_likelihood":
        sys.__excepthook__(exctype, value, traceback)
        print("\nFIGAROException: you probably have a sample that falls outside the given boundaries\n")
    else:
        sys.__excepthook__(exctype, value, traceback)



import numpy as np
from figaro.cumulative import fast_log_cumulative

def FindNearest_Volume(ra, dec, dist, value):
    """
    Find the pixel that contains the triplet (ra', dec', D') stored in value.
    
    Arguments:
        np.ndarray ra:   right ascension values used to build the grid
        np.ndarray dec:  declination values used to build the grid
        np.ndarray dist: luminosity distance values used to build the grid
        iterable value:  triplet to locate (ra', dec', D')
    
    Returns:
        np.ndarray: grid indices of pixel
    """
    idx = np.zeros(3, dtype = int)
    for i, (d, v) in enumerate(zip([ra, dec, dist], value)):
        idx[i] = int(np.abs(d-v).argmin())
    return idx

def FindNearest_Grid(grid, value):
    """
    Find the closest grid point to value.
    
    Arguments:
        np.ndarray grid: grid points (N_pts, N_dim), as with figaro.utils.recursive_grid
        iterable value:  value to locate
    
    Returns:
        np.ndarray: grid index
    """
    return abs(np.sum((grid - value)**2, axis = -1)).argmin()

def FindHeights(args):
    """
    Find height correspinding to a certain credible level given a sorted array of probabilities and the corresponding cumulative
    
    Arguments:
        tuple args: tuple containing the sorted array, the cumulative array and a double corresponding to the credible level
    
    Returns:
        double: height corresponding to the credible level
    """
    (sortarr,cumarr,level) = args
    return sortarr[np.abs(cumarr-np.log(level)).argmin()]

def FindHeightForLevel(inLogArr, adLevels, logdd):
    """
    Given a probability array, computes the heights corresponding to some given credible levels.
    
    Arguments:
        np.ndarray inLogArr: probability array
        iterable adLevels:   credible levels
        double logdd:        variables log differential (∑ log(dx_i))
        
    Returns:
        np.ndarray: heights corresponding to adLevels
    """
    # flatten and create reversed sorted list
    adSorted = np.ascontiguousarray(np.sort(inLogArr.flatten())[::-1])
    # create a normalized cumulative distribution
    adCum = fast_log_cumulative(adSorted + logdd)
    # find values closest to levels
    adHeights = []
    adLevels = np.ravel([adLevels])
    for level in adLevels:
        idx = (np.abs(adCum-np.log(level))).argmin()
        adHeights.append(adSorted[idx])
    adHeights = np.array(adHeights)
    return adHeights

def FindLevelForHeight(inLogArr, logvalue, logdd):
    """
    Given a probability array, computes the credible levels corresponding to a given height.
    
    Arguments:
        np.ndarray inLogArr: log probability array
        double logvalue:     height
        double logdd:        variables log differential (∑ log(dx_i))
    
    Returns:
        np.ndarray: credible level corresponding to logvalue
    """
    # flatten and create reversed sorted list
    adSorted = np.ascontiguousarray(np.sort(inLogArr.flatten())[::-1])
    # create a normalized cumulative distribution
    adCum = fast_log_cumulative(adSorted + logdd)
    # find index closest to value
    idx = (np.abs(adSorted-logvalue)).argmin()
    return np.exp(adCum[idx])

def ConfidenceVolume(log_volume_map, ra_grid, dec_grid, distance_grid, log_measure = None, adLevels = [0.50, 0.90]):
    """
    Compute the credible volume(s) for a 3D probability distribution
    
    Arguments:
        np.ndarray log_volume_map: probability density for each pixel
        np.ndarray ra_grid:        right ascension values used to build the grid
        np.ndarray dec_grid:       declination values used to build the grid
        np.ndarray distance_grid:  luminosity distance values used to build the grid
        iterable adLevels:         credible level(s)
    
    Returns:
        np.ndarray: credible volume(s)
        iterable:   indices of pixels within credible volume(s)
        np.ndarray: height(s) corresponding to credible volume(s)
    """
    dd  = np.diff(distance_grid)[0]
    ddec = np.diff(dec_grid)[0]
    dra = np.diff(ra_grid)[0]
    # create a normalized cumulative distribution
    log_volume_map_sorted = np.ascontiguousarray(np.sort(log_volume_map.flatten())[::-1])
    if log_measure is not None:
        log_measure_sorted = np.ascontiguousarray(log_measure.flatten()[np.argsort(log_volume_map.flatten())][::-1])
    else:
        log_measure_sorted = np.zeros(log_volume_map_sorted.shape)
    log_norm              = fast_log_cumulative(log_volume_map_sorted + log_measure_sorted + np.log(dra) + np.log(ddec) + np.log(dd))[-1]
    log_volume_map_sorted = log_volume_map_sorted - log_norm
    log_volume_map        = log_volume_map - log_norm
    log_volume_map_cum    = fast_log_cumulative(log_volume_map_sorted + log_measure_sorted + np.log(dra) + np.log(ddec) + np.log(dd))
    # find the indeces  corresponding to the given CLs
    adLevels = np.ravel([adLevels])
    args = [(log_volume_map_sorted, log_volume_map_cum, level) for level in adLevels]
    adHeights = [FindHeights(a) for a in args]
    volumes         = []
    index           = []
    for height in adHeights:
        (i_ra, i_dec, i_d,) = np.where(log_volume_map>=height)
        volumes.append(np.sum([distance_grid[i_d]**2. *np.cos(dec_grid[i_dec]) * dd * dra * ddec for i_d,i_dec in zip(i_d,i_dec)]))
        index.append(np.array([i_ra, i_dec, i_d]).T)
    volume_confidence = np.array(volumes)
    
    return volume_confidence, index, np.array(adHeights)

def ConfidenceArea(log_skymap, ra_grid, dec_grid, log_measure = None, adLevels = [0.50, 0.90]):
    """
    Compute the credible area(s) for a 2D probability distribution
    
    Arguments:
        np.ndarray log_skymap: probability density for each pixel
        np.ndarray ra_grid:    right ascension values used to build the grid
        np.ndarray dec_grid:   declination values used to build the grid
        iterable adLevels:     credible level(s)
    
    Returns:
        np.ndarray: credible area(s)
        iterable:   indices of pixels within credible area(s)
        np.ndarray: height(s) corresponding to credible area(s)
    """
    # create a normalized cumulative distribution
    ddec = np.diff(dec_grid)[0]
    dra = np.diff(ra_grid)[0]
    log_skymap_sorted  = np.ascontiguousarray(np.sort(log_skymap.flatten())[::-1])
    if log_measure is not None:
        log_measure_sorted = np.ascontiguousarray(log_measure.flatten()[np.argsort(log_skymap.flatten())][::-1])
    else:
        log_measure_sorted = np.zeros(log_skymap_sorted.shape)
    log_norm           = fast_log_cumulative(log_skymap_sorted + log_measure_sorted.flatten() + np.log(dra) + np.log(ddec))[-1]
    log_skymap         = log_skymap - log_norm
    log_skymap_sorted  = log_skymap_sorted - log_norm
    log_skymap_cum     = fast_log_cumulative(log_skymap_sorted + log_measure_sorted.flatten() + np.log(dra) + np.log(ddec) - log_norm)
    # find the indeces  corresponding to the given CLs
    adLevels = np.ravel([adLevels])
    args = [(log_skymap_sorted, log_skymap_cum, level) for level in adLevels]
    adHeights = [FindHeights(a) for a in args]
    areas = []
    index = []
    for height in adHeights:
        (i_ra,i_dec,) = np.where(log_skymap>=height)
        areas.append(np.sum([dra*np.cos(dec_grid[i_d])*ddec for i_d in i_dec])*(180.0/np.pi)**2.0)
        index.append(np.array([i_ra, i_dec]).T)
    area_confidence = np.array(areas)
    
    return area_confidence, index, np.array(adHeights)




from figaro.transform import transform_from_probit, transform_to_probit

"""
A gentle introduction to decorators: https://towardsdatascience.com/how-to-use-decorators-in-python-by-example-b398328163b
"""

def antiprobit(func):
    """
    Transform a point x from probit space to natural space and returns the function evaluated at the natural point y
    """
    def f_transf(ref, x, *args, **kwargs):
        if not ref.probit:
            return func(ref, x, *args)
        y = transform_from_probit(x, ref.bounds)
        return func(ref, y, *args)
    return f_transf

def probit(func):
    """
    Transform a point x from natural space to probit space and returns the function evaluated at the probit point y
    """
    def f_transf(ref, x, *args, **kwargs):
        if not ref.probit:
            return func(ref, x, *args)
        y = transform_to_probit(x, ref.bounds)
        return func(ref, y, *args)
    return f_transf

def from_probit(func):
    """
    Evaluate a function that samples points in probit space and return these points after transforming them to natural space
    """
    def f_transf(ref, *args, **kwargs):
        if not ref.probit:
            return func(ref, *args)
        y = func(ref, *args)
        return transform_from_probit(y, ref.bounds)
    return f_transf



import numpy as np
import h5py
import warnings
import json
import dill
import copy
import importlib
from figaro.exceptions import FIGAROException
from figaro.mixture import mixture
from figaro.cosmology import Planck18, Planck15
from pathlib import Path
from tqdm import tqdm

supported_extensions = ['h5', 'hdf5', 'txt', 'dat', 'csv']
supported_waveforms  = ['combined', 'imr', 'seob']
injected_pars        = ['m1', 'm2', 'z', 's1x', 's1y', 's1z', 's2x', 's2y', 's2z', 'ra', 'dec']
loadable_inj_pars    = injected_pars + ['q', 'chi_eff', 'chi_p', 's1', 's2']
mass_parameters      = ['m1', 'm2', 'm1_detect', 'm2_detect', 'mc', 'mt', 'q']
spin_parameters      = ['s1x', 's1y', 's1z', 's2x', 's2y', 's2z', 's1', 's2', 'chi_eff', 'chi_p']

GW_par = {'m1'                 : 'mass_1_source',
          'm2'                 : 'mass_2_source',
          'm1_detect'          : 'mass_1',
          'm2_detect'          : 'mass_2',
          'mc'                 : 'chirp_mass_source',
          'mc_detect'          : 'chirp_mass',
          'mt'                 : 'total_mass_source',
          'z'                  : 'redshift',
          'q'                  : 'mass_ratio',
          'eta'                : 'symmetric_mass_ratio',
          'chi_eff'            : 'chi_eff',
          'chi_p'              : 'chi_p',
          'ra'                 : 'ra',
          'dec'                : 'dec',
          'luminosity_distance': 'luminosity_distance',
          'cos_theta_jn'       : 'cos_theta_jn',
          'cos_tilt_1'         : 'cos_tilt_1',
          'cos_tilt_2'         : 'cos_tilt_2',
          's1x'                : 'spin_1x',
          's2x'                : 'spin_2x',
          's1y'                : 'spin_1y',
          's2y'                : 'spin_2y',
          's1z'                : 'spin_1z',
          's2z'                : 'spin_2z',
          's1'                 : 'spin_1',
          's2'                 : 'spin_2',
          'psi'                : 'psi',
          'cos_iota'           : 'cos_iota',
          'phase'              : 'phase',
          'tc'                 : 'geocent_time',
          'snr'                : 'network_matched_filter_snr',
          'far'                : 'far',
          'log_prior'          : 'log_prior',
          'log_likelihood'     : 'log_likelihood',
          }

# LVK 03 injections have a slightly different nomenclature (no underscore)
# See https://zenodo.org/records/7890437
inj_par = copy.deepcopy(GW_par)
inj_par['m1']                  = 'mass1_source'
inj_par['m2']                  = 'mass2_source'
inj_par['m1_detect']           = 'mass1'
inj_par['m2_detect']           = 'mass2'
inj_par['s1x']                 = 'spin1x'
inj_par['s1y']                 = 'spin1y'
inj_par['s1z']                 = 'spin1z'
inj_par['s2x']                 = 'spin2x'
inj_par['s2y']                 = 'spin2y'
inj_par['s2z']                 = 'spin2z'
inj_par['luminosity_distance'] = 'distance'

supported_pars = [p for p in GW_par.keys() if not p in ['snr', 'far']]

def available_gw_pars():
    """
    Print a list of available GW parameters
    """
    print(supported_pars)

def load_single_event(event, seed = False, par = None, n_samples = -1, cosmology = 'Planck18', volume = False, waveform = 'combined', snr_threshold = None, far_threshold = None):
    '''
    Loads the data from .txt/.dat files (for simulations) or .h5/.hdf5 files (posteriors from GWTC) for a single event.
    Default cosmological parameters from Planck Collaboration (2021) in a flat Universe (https://www.aanda.org/articles/aa/pdf/2020/09/aa33910-18.pdf)
    Not all GW parameters are implemented: run figaro.load.available_gw_pars() for a list of the available parameters.
    
    Arguments:
        str or Path file:     file with samples
        bool seed:            fixes the seed to a default value (1) for reproducibility
        list-of-str par:      list with parameter(s) to extract from GW posteriors (m1, m2, mc, z, chi_effective)
        int n_samples:        number of samples for (random) downsampling. Default -1: all samples
        str cosmology:        set of cosmological parameters (Planck18 or Planck15)
        bool volume:          if True, loads RA, dec and Luminosity distance (for skymaps)
        str waveform:         waveform family to be used ('combined', 'seob', 'imr')
        double snr_threhsold: SNR threshold for event filtering. For injection analysis only.
        double far_threshold: FAR threshold for event filtering. For injection analysis only.
        
    Returns:
        np.ndarray: samples
        np.ndarray: name
    '''
    if seed:
        rdstate = np.random.RandomState(seed = 1)
    else:
        rdstate = np.random.RandomState()
    event = Path(event).resolve()
    name, ext = str(event).split('/')[-1].split('.')
    if volume:
        par = ['ra', 'dec', 'luminosity_distance']
    if not ext in supported_extensions:
        raise TypeError("File {0}.{1} is not supported".format(name, ext))
    if not ext in ['h5', 'hdf5']:
        if par is not None:
            warnings.warn("Par names (or volume keyword) are ignored for .txt/.dat/.csv files")
        if n_samples > -1:
            samples = np.atleast_1d(np.loadtxt(event))
            s = int(min([n_samples, len(samples)]))
            out = samples[rdstate.choice(np.arange(len(samples)), size = s, replace = False)]
        else:
            out = np.loadtxt(event)
    else:
        # Check that a list of parameters is passed
        if par is None:
            raise TypeError("Please provide a list of parameter names you want to load (e.g. ['m1']).")
        # Check that all the parametes are loadable
        if not np.all([p in GW_par.keys() for p in par]):
            wrong_pars = [p for p in par if not p in GW_par.keys()]
            raise FIGAROException("The following parameters are not implemented: "+", ".join(wrong_pars)+". Run figaro.load.available_gw_pars() for a list of available parameters.")
        # If everything is ok, load the samples
        else:
            out = _unpack_gw_posterior(event, par = par, n_samples = n_samples, cosmology = cosmology, rdstate = rdstate, waveform = waveform, snr_threshold = snr_threshold, far_threshold = far_threshold)
    
    if out is None:
        return out, name
    
    if len(np.shape(out)) == 1:
        out = np.atleast_2d(out).T
    return out, name

def load_data(path, seed = False, par = None, n_samples = -1, cosmology = 'Planck18', volume = False, waveform = 'combined', snr_threshold = None, far_threshold = None, verbose = True):
    '''
    Loads the data from .txt files (for simulations) or .h5/.hdf5/.dat files (posteriors from GWTC-x).
    Default cosmological parameters from Planck Collaboration (2021) in a flat Universe (https://www.aanda.org/articles/aa/pdf/2020/09/aa33910-18.pdf)
    Not all GW parameters are implemented: run figaro.load.available_gw_pars() for a list of available parameters.
    
    Arguments:
        str or Path path:     folder with data files
        bool seed:            fixes the seed to a default value (1) for reproducibility
        list-of-str par:      list with parameter(s) to extract from GW posteriors
        int n_samples:        number of samples for (random) downsampling. Default -1: all samples
        str cosmology:        set of cosmological parameters (Planck18 or Planck15)
        str waveform:         waveform family to be used ('combined', 'seob', 'imr')
        double snr_threhsold: SNR threshold for event filtering. For injection analysis only.
        double far_threshold: FAR threshold for event filtering. For injection analysis only.
        bool verbose:         show progress bar

    Returns:
        np.ndarray: samples
        np.ndarray: names
    '''
    folder      = Path(path).resolve()
    event_files = [Path(folder,f) for f in folder.glob('[!.]*')] # Ignores hidden files
    events      = []
    names       = []
    n_events    = len(event_files)
    removed_snr = False
    if volume:
        par = ['ra', 'dec', 'luminosity_distance']
    if n_events == 0:
        raise FIGAROException("Empty folder")
    for event in tqdm(event_files, desc = 'Loading events', disable = not(verbose)):
        if seed:
            rdstate = np.random.RandomState(seed = 1)
        else:
            rdstate = np.random.RandomState()
        name, ext = str(event).split('/')[-1].split('.')
        names.append(name)
        if not ext in supported_extensions:
            raise TypeError("File {0}.{1} is not supported".format(name, ext))
        if not ext in ['h5', 'hdf5']:
            if par is not None:
                warnings.warn("Par names (or volume keyword) are ignored for .txt/.dat/.csv files")
            if n_samples > -1:
                samples = np.atleast_1d(np.loadtxt(event))
                s = int(min([n_samples, len(samples)]))
                samples_subset = samples[rdstate.choice(np.arange(len(samples)), size = s, replace = False)]
                if len(np.shape(samples_subset)) == 1:
                    samples_subset = np.atleast_2d(samples_subset).T
                events.append(samples_subset)
                    
            else:
                samples = np.atleast_1d(np.loadtxt(event))
                if len(np.shape(samples)) == 1:
                    samples = np.atleast_2d(samples).T
                events.append(samples)
                
        else:
            # Check that a list of parameters is passed
            if par is None:
                raise TypeError("Please provide a list of parameter names you want to load (e.g. ['m1']).")
            # Check that all the parametes are loadable
            if not np.all([p in GW_par.keys() for p in par]):
                wrong_pars = [p for p in par if not p in GW_par.keys()]
                raise FIGAROException("The following parameters are not implemented: "+", ".join(wrong_pars)+". Run figaro.load.available_gw_pars() for a list of available parameters.")
            # If everything is ok, load the samples
            else:
                out = _unpack_gw_posterior(event, par = par, n_samples = n_samples, cosmology = cosmology, rdstate = rdstate, waveform = waveform, snr_threshold = snr_threshold, far_threshold = far_threshold)
                if out is not None:
                    if out.shape[-1] == len(par):
                        events.append(out)
                    elif 'snr' in par:
                        removed_snr = True
                        names.remove(name)
                else:
                    names.remove(name)
    if removed_snr:
        warnings.warn("At least one event does not have SNR samples. These events cannot be loaded for this parameter choices.")
    return (events, np.array(names))

def _unpack_gw_posterior(event, par, cosmology, rdstate, n_samples = -1, waveform = 'combined', snr_threshold = None, far_threshold = None):
    '''
    Reads data from .h5/.hdf5 GW posterior files.
    For GWTC-3 data release, it uses by default the Mixed posterior samples.
    Not all parameters are implemented: run figaro.load.available_gw_pars() for a list of available parameters.
    The waveform argument allows the user to select a waveform family. The default value, 'combined' uses samples from both imr and seob waveforms.
    For SEOB waveforms, the following waveform models are used (in descending priority order):
        * SEOBNRv4PHM
        * SEOBNRv4P
        * SEOBNRv4
    For IMR waveforms, in descending order:
        * IMRPhenomXPHM
        * IMRPhenomPv2
        * IMRPhenomPv3HM
    
    Arguments:
        str event:            file to read
        str par:              parameter to extract
        tuple cosmology:      cosmological parameters (h, om, ol)
        int n_samples:        number of samples for (random) downsampling. Default -1: all samples
        str waveform:         waveform family to be used ('combined', 'imr', 'seob')
        double snr_threhsold: SNR threshold for event filtering. For injection analysis only.
        double far_threshold: FAR threshold for event filtering. For injection analysis only.
    
    Returns:
        np.ndarray: samples
    '''
    if cosmology == 'Planck18':
        omega = Planck18
    elif cosmology == 'Planck15':
        omega = Planck15
    else:
        raise FIGAROException("Cosmology not supported")
    if not waveform in supported_waveforms:
        raise FIGAROException("Unknown waveform: please use 'combined' (default), 'imr' or 'seob'")
    
    if far_threshold is not None and snr_threshold is not None:
        warnings.warn("Both FAR and SNR threshold provided. FAR will be used.")
        snr_threshold = None
    
    if far_threshold is not None:
        if not 'far' in par:
            par = np.append(par, 'far')
    elif snr_threshold is not None:
        if not 'snr' in par:
            par = np.append(par, 'snr')
    
    with h5py.File(Path(event), 'r') as f:
        samples     = []
        loaded_pars = []
        flag_filter = False
        try:
            try:
                # LVK injections
                try:
                    data = f['MDC']['posterior_samples']
                # Playground
                except:
                    data = f['posterior_samples']
                MDC_flag = True
            # GWTC-2, GWTC-3
            except:
                MDC_flag = False
                if waveform == 'combined':
                    # GWTC-2
                    try:
                        data = f['PublicationSamples']['posterior_samples']
                    # GWTC-3
                    except KeyError:
                        try:
                            data = f['C01:Mixed']['posterior_samples']
                        except:
                            try:
                                data = f['IMRPhenomXPHM']['posterior_samples']
                            except:
                                data = f['SEOBNRv4PHM']['posterior_samples']
                else:
                    if waveform == 'imr':
                        try:
                            try:
                                data = f['C01:IMRPhenomXPHM']['posterior_samples']
                            except:
                                data = f['IMRPhenomXPHM']['posterior_samples']
                        except:
                            try:
                                try:
                                    data = f['C01:IMRPhenomPv2']['posterior_samples']
                                except:
                                    data = f['IMRPhenomPv2']['posterior_samples']
                            except:
                                try:
                                    try:
                                        data = f['C01:IMRPhenomPv3HM']['posterior_samples']
                                    except:
                                        data = f['IMRPhenomPv3HM']['posterior_samples']
                                except:
                                    try:
                                        try:
                                            data = f['C01:IMRPhenomXPHM:LowSpin']['posterior_samples']
                                        except:
                                            data = f['IMRPhenomXPHM:LowSpin']['posterior_samples']
                                    except:
                                        try:
                                            data = f['C01:IMRPhenomPv2_NRTidal-LS']['posterior_samples']
                                        except:
                                            data = f['IMRPhenomPv2_NRTidal-LS']['posterior_samples']
                    if waveform == 'seob':
                        try:
                            try:
                                data = f['C01:SEOBNRv4PHM']['posterior_samples']
                            except:
                                data = f['SEOBNRv4PHM']['posterior_samples']
                        except:
                            try:
                                try:
                                    data = f['C01:SEOBNRv4P']['posterior_samples']
                                except:
                                    data = f['SEOBNRv4P']['posterior_samples']
                            except:
                                try:
                                    try:
                                        data = f['C01:SEOBNRv4']['posterior_samples']
                                    except:
                                        data = f['SEOBNRv4']['posterior_samples']
                                except:
                                    try:
                                        data = f['C01:SEOBNRv4T_surrogate_LS']['posterior_samples']
                                    except:
                                        data = f['SEOBNRv4T_surrogate_LS']['posterior_samples']
                
            for name, lab in zip(GW_par.keys(), GW_par.values()):
                if name in par:
                    if name == 'snr':
                        try:
                            if MDC_flag or waveform != 'combined':
                                snr = np.array(data[lab])
                                samples.append(data[lab])
                            if snr_threshold is not None:
                                flag_filter = True
                        except:
                            warnings.warn("SNR filter is not available with this dataset.")
                    elif name == 'far' and far_threshold is not None:
                        try:
                            flag_filter = True
                            far = np.array(data[lab])
                        except:
                            warnings.warn("FAR filter is not available with this dataset.")
                    elif name == 's1':
                        try:
                            samples.append(data[lab])
                        except:
                            samples.append(np.sqrt(data['spin_1x']**2+data['spin_1y']**2+data['spin_1z']**2))
                    elif name == 's2':
                        try:
                            samples.append(data[lab])
                        except:
                            samples.append(np.sqrt(data['spin_2x']**2+data['spin_2y']**2+data['spin_2z']**2))
                    elif name == 'luminosity_distance':
                        try:
                            samples.append(data[lab])
                        except:
                            samples.append(np.exp(data['logdistance']))
                    else:
                        samples.append(data[lab])
                    loaded_pars.append(name)
            
            if len(par) == 1:
                samples = np.atleast_2d(samples).T
            else:
                par = np.array(par)
                loaded_pars = np.array(loaded_pars)
                samples_loaded = np.array(samples)
                samples = []
                for pi in par:
                    if not (pi == 'far' or (pi == 'snr' and flag_filter)):
                        samples.append(samples_loaded[np.where(loaded_pars == pi)[0]].flatten())
                samples = np.array(samples)
                if flag_filter:
                    if far_threshold is not None:
                        samples = samples[:, np.where((far < far_threshold) & (far > 0))[0]]
                    if snr_threshold is not None:
                        samples = samples[:, np.where(snr > snr_threshold)[0]]
                samples = samples.T
            if n_samples > -1:
                s = int(min([n_samples, len(samples)]))
                return samples[rdstate.choice(np.arange(len(samples)), size = s, replace = False)]
            else:
                return samples

        # GWTC-1
        except KeyError:
            if waveform == 'combined':
                label = 'Overall_posterior'
            elif waveform == 'imr':
                label = 'IMRPhenomPv2_posterior'
            elif waveform == 'seob':
                label = 'SEOBNRv3_posterior'
            try:
                data = f[label]
            except KeyError:
                try:
                    # GW170817
                    data = f['IMRPhenomPv2NRT_lowSpin_posterior']
                except:
                    print("Skipped event {0} (not loadable yet)".format(Path(event).parts[-1].split('.')[0]))
                    return None
            
            # Provided quantities
            names_GWTC1 = {'m1_detect'          : 'm1_detector_frame_Msun',
                           'm2_detect'          : 'm2_detector_frame_Msun',
                           'ra'                 : 'right_ascension',
                           'dec'                : 'declination',
                           'luminosity_distance': 'luminosity_distance_Mpc',
                           'cos_theta_jn'       : 'costheta_jn',
                           's1'                 : 'spin1',
                           's2'                 : 'spin2',
                           'cos_tilt_1'         : 'costilt1',
                           'cos_tilt_2'         : 'costilt2',
                           }
            
            ss = {name: data[lab] for name, lab in zip(names_GWTC1.keys(), names_GWTC1.values())}

            # Derived quantities
            ss['z']       = omega.Redshift(ss['luminosity_distance'])
            ss['m1']      = ss['m1_detect']/(1+ss['z'])
            ss['m2']      = ss['m2_detect']/(1+ss['z'])
            ss['mc']      = (ss['m1']*ss['m2'])**(3./5.)/(ss['m1']+ss['m2'])**(1./5.)
            ss['mt']      = ss['m1']+ss['m2']
            ss['q']       = ss['m2']/ss['m1']
            ss['eta']     = ss['m1']*ss['m2']/(ss['m1']+ss['m2'])**2
            ss['s1z']     = ss['s1']*ss['cos_tilt_1']
            ss['s2z']     = ss['s2']*ss['cos_tilt_2']
            ss['chi_eff'] = (ss['s1z'] + ss['q']*ss['s2z'])/(1+ss['q'])

            for name in GW_par.keys():
                if name in par:
                    if not (name == 'snr' or name == 'far'):
                        samples.append(ss[name])
                        loaded_pars.append(name)

            if len(par) == 1:
                samples = np.atleast_2d(samples).T
            else:
                par = np.array(par)
                loaded_pars = np.array(loaded_pars)
                samples_loaded = np.array(samples)
                samples = []
                for pi in par:
                    if not (pi == 'snr' or pi == 'far'):
                        samples.append(samples_loaded[np.where(loaded_pars == pi)[0]].flatten())
                samples = np.array(samples)
                samples = samples.T
            if n_samples > -1:
                s = int(min([n_samples, len(samples)]))
                return samples[rdstate.choice(np.arange(len(samples)), size = s, replace = False)]
            else:
                return samples

def save_density(draws, folder = '.', name = 'density', ext = 'json'):
    """
    Exports a list of figaro.mixture instances to file

    Arguments:
        list draws:         list of mixtures to be saved
        str or Path folder: folder in which the output file will be saved
        str name:           name to be given to output file
        str ext:            file extension (pkl or json)
    """
    if ext == 'pkl':
        with open(Path(folder, name+'.pkl'), 'wb') as f:
            dill.dump(draws, f)
    elif ext == 'json':
        if len(np.shape(draws)) == 1:
            draws = np.atleast_2d(draws)
        ll = []
        for draws_i in draws:
            list_of_dicts = [dr.__dict__.copy() for dr in draws_i]
            for density in list_of_dicts:
                density.pop('components')
                for key in density.keys():
                    value = density[key]
                    if isinstance(value, np.ndarray):
                        value = value.tolist()
                    else:
                        value = float(value)
                    density[key] = value
            ll.append(list_of_dicts)
        s = json.dumps(ll)
        with open(Path(folder, name + '.json'), 'w') as f:
            json.dump(s, f)
    else:
        raise FIGAROException("Extension {0} is not supported. Valid extensions are pkl or json.")

def load_density(path, make_comp = True):
    """
    Loads a list of figaro.mixture instances from path.
    If the requested file extension (pkl or json) is not available, it tries loading the other.

    Arguments:
        str or Path path: path with draws (file or folder)
        bool make_comp:   make component objects

    Returns:
        list: figaro.mixture object instances
    """
    path = Path(path)
    if path.is_file():
        return _load_density_file(path, make_comp)
    else:
        files = [_load_density_file(file, make_comp) for file in path.glob('*.[jp][sk][ol]*') if not file.stem == 'posteriors_single_event']
        if len(files) > 0:
            return files
        else:
            raise FIGAROException("Density file(s) not found")

def _load_density_file(file, make_comp = True):
    """
    Loads a list of figaro.mixture instances from file.
    If the requested file extension (pkl or json) is not available, it tries loading the other.

    Arguments:
        str or Path file: file with draws
        bool make_comp:   make component objects

    Returns
        list: figaro.mixture object instances
    """
    file = Path(file)
    ext  = file.suffix
    if ext == '.pkl':
        try:
            return _load_pkl(file)
        except FileNotFoundError:
            try:
                return _load_json(file.with_suffix('.json'), make_comp)
            except FileNotFoundError:
                raise FIGAROException("{0} not found. Please provide it or re-run the inference.".format(file.name))
    elif ext == '.json':
        try:
            return _load_json(file, make_comp)
        except FileNotFoundError:
            try:
                return _load_pkl(file.with_suffix('.pkl'))
            except FileNotFoundError:
                raise FIGAROException("{0} not found. Please provide it or re-run the inference.".format(file.name))
    else:
        raise FIGAROException("Extension {0} is not supported. Please provide .pkl or .json file.".format(file.suffix))

def _load_pkl(file):
    """
    Loads a list of figaro.mixture instances from pkl file

    Arguments:
        str or Path file: file with draws

    Returns
        list: figaro.mixture object instances
    """
    with open(file, 'rb') as f:
        draws = dill.load(f)
    return draws

def _load_json(file, make_comp = True):
    """
    Loads a list of figaro.mixture instances from json file

    Arguments:
        str or Path file: file with draws
        bool make_comp:   make component objects

    Returns
        list: figaro.mixture object instances
    """
    with open(Path(file), 'r') as fjson:
        dictjson = json.loads(json.load(fjson))
    ll = []
    for list_of_dict in dictjson:
        draws = []
        for dict_ in list_of_dict:
            dict_.pop('log_w')
            for key in dict_.keys():
                value = dict_[key]
                if isinstance(value, list):
                    dict_[key] = np.array(value)
                if key == 'probit':
                    dict_[key] = bool(value)
            draws.append(mixture(**dict_, make_comp = make_comp))
        ll.append(draws)
    if len(ll) == 1:
        return ll[0]
    return ll

def load_selection_function(file, par = None, far_threshold = 1):
    """
    Loads the selection function, either from a python module containing a method called 'selection_function' or via injections.
    If injections, it assumes that the last column of a txt/csv/dat file contains the sampling pdf.
    
    Arguments:
        str or Path file: selection function file
        list-of-str par:  list with parameter(s) to extract from GW posteriors
        double threshold: FAR threshold to filter LVK injections
    
    Returns:
        np.ndarray or callable: detected samples or callable with approximant
        np.ndarray or NoneType: injection pdf (for samples) or None (for approximant)
        int or NoneType:        total number of injections
        double duration:        duration of the observation
    """
    file = Path(file)
    ext  = file.parts[-1].split('.')[1]
    if not ext in supported_extensions + ['py']:
        raise FIGAROException("Selection function file not supported")
    if ext == 'py':
        selfunc_file_name = file.parts[-1].split('.')[0]
        spec              = importlib.util.spec_from_file_location(selfunc_file_name, file)
        selfunc_module    = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(selfunc_module)
        selfunc           = selfunc_module.selection_function
        inj_pdf           = None
        n_total_inj       = None
        try:
            duration      = selfunc_module.duration
        except:
            duration      = 1.
    else:
        if not ext in ['h5','hdf5']:
            samples     = np.loadtxt(file)
            det_idx     = samples[:,-1]
            selfunc     = samples[:,:-2][det_idx == 1]
            inj_pdf     = samples[:,-2][det_idx == 1]
            n_total_inj = len(samples)
            duration    = 1.
        else:
            selfunc, inj_pdf, n_total_inj, duration = _unpack_injections(file, par, far_threshold)
    return selfunc, inj_pdf, n_total_inj, duration

def _unpack_injections(file, par, far_threshold = 1.):
    """
    Reads data from .h5/.hdf5 injection file (https://zenodo.org/records/7890437).
    A sample is considered detected if at least one of the searches calls a detection.
    
    Arguments:
        str event:            file to read
        str par:              parameter to extract
        double far_threshold: FAR threshold for injection filtering
    
    Returns:
        np.ndarray:      samples
        np.ndarray:      injection pdf
        int or NoneType: total number of injections
        double duration: duration of the observation
    """
    # Check that a list of parameters is passed
    if par is None:
        raise TypeError("Please provide a list of parameter names you want to load (e.g. ['m1']).")
    # Check that all the parametes are loadable
    if not np.all([p in loadable_inj_pars for p in par]):
        wrong_pars = [p for p in par if not p in loadable_inj_pars]
        raise FIGAROException("The following parameters are not implemented: "+", ".join(wrong_pars))
    with h5py.File(file, 'r') as f:
        data = f['injections']
        n_total_inj = int(data.attrs['total_generated'])
        duration    = data.attrs['analysis_time_s']/(60.*60.*24.*365) # Years
        # Detected injections
        far_cwb    = np.array(data['far_cwb'])
        far_gstlal = np.array(data['far_gstlal'])
        far_mbta   = np.array(data['far_mbta'])
        try:
            far_pycbc = np.array(data['far_pycbc_bbh'])
        except:
            far_pycbc = np.array(data['far_pycbc_hyperbank'])
        idx = np.where((far_cwb < far_threshold) | (far_gstlal < far_threshold) | (far_mbta < far_threshold) | (far_pycbc < far_threshold))[0]
        # Load samples
        samples = np.zeros((len(par), len(idx)))
        inj_pdf = np.ones(len(idx))
        # Parameters
        m1  = np.array(data[inj_par['m1']])[idx]
        m2  = np.array(data[inj_par['m2']])[idx]
        q   = m2/m1
        s1x = np.array(data[inj_par['s1x']])[idx]
        s1y = np.array(data[inj_par['s1y']])[idx]
        s1z = np.array(data[inj_par['s1z']])[idx]
        s2x = np.array(data[inj_par['s2x']])[idx]
        s2y = np.array(data[inj_par['s2y']])[idx]
        s2z = np.array(data[inj_par['s2z']])[idx]
        for i, lab in enumerate(par):
            # Already available parameters:
            if lab in injected_pars:
                if not lab == 'cos_theta_jn':
                    samples[i] = np.array(data[inj_par[lab]])[idx]
                else:
                    samples[i] = np.cos(np.array(data[inj_par[lab]])[idx])
            else:
                # Masses
                if lab == 'q':
                    samples[i] = q
                if lab == 'mt':
                    samples[i] = m1 + m2
                if lab == 'mc':
                    samples[i] = m1*m2**(3./5.)/(m1+m2)**(1./5.)
                # Spins
                if lab == 's1':
                    samples[i] = np.sqrt(s1x**2+s1y**2+s1z**2)
                if lab == 's2':
                    samples[i] = np.sqrt(s2x**2+s2y**2+s2z**2)
                if lab == 'chi_eff':
                    samples[i] = (s1z + q*s2z)/(1+q)
                if lab == 'chi_p':
                    samples[i] = np.maximum(np.sqrt(s1x**2+s1y**2), np.sqrt(s2x**2+s2y**2)*q*(4*q+3)/(4+3*q))
        # Sampling pdf
        n_mass_pars = len([lab for lab in par if lab in mass_parameters])
        n_spin_pars = len([lab for lab in par if lab in spin_parameters])
        # Masses
        if n_mass_pars == 1:
            inj_pdf *= np.array(data['mass1_source_sampling_pdf'])[idx]
        elif n_mass_pars == 2:
            inj_pdf *= np.array(data['mass1_source_mass2_source_sampling_pdf'])[idx]
        if 'q' in par:
            inj_pdf *= m1
        # Spins
        if n_spin_pars > 0:
            inj_pdf *= (np.array(data['spin1x_spin1y_spin1z_sampling_pdf'])[idx]*np.array(data['spin2x_spin2y_spin2z_sampling_pdf'])[idx])**(n_spin_pars/6.)
        if 's1' in par or 'chi_eff' in par or 'chi_p' in par:
            inj_pdf /= 2*np.pi*(s1x**2+s1y**2+s1z**2)
        if 's2' in par or 'chi_eff' in par or 'chi_p' in par:
            inj_pdf /= 2*np.pi*(s2x**2+s2y**2+s2z**2)
        # Distance
        if 'z' in par:
            inj_pdf *= np.array(data['redshift_sampling_pdf'])[idx]
        # Sky position
        if 'ra' in par:
            inj_pdf *= np.array(data['right_ascension_sampling_pdf'])[idx]
        if 'dec' in par:
            inj_pdf *= np.array(data['declination_sampling_pdf'])[idx]
    return samples.T, inj_pdf, n_total_inj, duration



import numpy as np
from numba import njit, prange
from figaro._numba_functions import log_add

@njit
def fast_log_cumulative(f):
    """
    Compute log cdf of probability density f
    
    Arguments:
        :np.ndarray f: probability density
    
    Returns:
        :np.ndarray: log cdf
    """
    n = f.shape[0]
    h = np.zeros(n, dtype = np.double)
    h[0] = f[0]
    for i in prange(1,n):
        h[i] = log_add(h[i-1],f[i])
    return h

@njit
def fast_cumulative(f):
    """
    Compute cdf of probability density f
    
    Arguments:
        :np.ndarray f: probability density
    
    Returns:
        :np.ndarray: cdf
    """
    n = f.shape[0]
    h = np.zeros(n, dtype = np.double)
    h[0] = f[0]
    for i in prange(1,n):
        h[i] = h[i-1] + f[i]
    return h



import numpy as np

import optparse
import dill
import importlib

from pathlib import Path
from tqdm import tqdm
from warnings import warn

from figaro.mixture import DPGMM
from figaro.utils import save_options, load_options, get_priors
from figaro.plot import plot_median_cr, plot_multidim
from figaro.load import load_single_event, load_selection_function, save_density, load_density, supported_extensions, supported_pars

def main():

    parser = optparse.OptionParser(prog = 'figaro-density', description = 'Probability density reconstruction')
    # Input/output
    parser.add_option("-i", "--input", type = "string", dest = "input", help = "File with samples", default = None)
    parser.add_option("-b", "--bounds", type = "string", dest = "bounds", help = "Density bounds. Must be a string formatted as '[[xmin, xmax], [ymin, ymax],...]'. For 1D distributions use '[xmin, xmax]'. Quotation marks are required and scientific notation is accepted", default = None)
    parser.add_option("-o", "--output", type = "string", dest = "output", help = "Output folder. Default: same directory as samples", default = None)
    parser.add_option("--ext", dest = "ext", type = "choice", choices = ['pkl', 'json'], help = "Format of mixture output file", default = 'json')
    parser.add_option("--inj_density", type = "string", dest = "inj_density_file", help = "Python module with injected density - please name the method 'density'", default = None)
    parser.add_option("--selfunc", type = "string", dest = "selfunc_file", help = "Python module with selection function - please name the method 'selection_function'", default = None)
    parser.add_option("--parameter", type = "string", dest = "par", help = "GW parameter(s) to be read from file", default = None)
    parser.add_option("--waveform", type = "choice", dest = "wf", help = "Waveform to load from samples file. To be used in combination with --parameter.", choices = ['combined', 'seob', 'imr'], default = 'combined')
    # Plot
    parser.add_option("-p", "--postprocess", dest = "postprocess", action = 'store_true', help = "Postprocessing", default = False)
    parser.add_option("--symbol", type = "string", dest = "symbol", help = "LaTeX-style quantity symbol, for plotting purposes", default = None)
    parser.add_option("--unit", type = "string", dest = "unit", help = "LaTeX-style quantity unit, for plotting purposes", default = None)
    # Settings
    parser.add_option("--draws", type = "int", dest = "draws", help = "Number of draws", default = 100)
    parser.add_option("--n_samples_dsp", type = "int", dest = "n_samples_dsp", help = "Number of samples to analyse (downsampling). Default: all", default = -1)
    parser.add_option("--exclude_points", dest = "exclude_points", action = 'store_true', help = "Exclude points outside bounds from analysis", default = False)
    parser.add_option("--cosmology", type = "choice", dest = "cosmology", help = "Set of cosmological parameters. Default values from Planck (2021)", choices = ['Planck18', 'Planck15'], default = 'Planck18')
    parser.add_option("--sigma_prior", dest = "sigma_prior", type = "string", help = "Expected standard deviation (prior) - single value or n-dim values. If None, it is estimated from samples", default = None)
    parser.add_option("--fraction", dest = "fraction", type = "float", help = "Fraction of samples standard deviation for sigma prior. Overrided by sigma_prior.", default = None)
    parser.add_option("--snr_threshold", dest = "snr_threshold", type = "float", help = "SNR threshold for simulated GW datasets", default = None)
    parser.add_option("--far_threshold", dest = "far_threshold", type = "float", help = "FAR threshold for simulated GW datasets", default = None)
    parser.add_option("--no_probit", dest = "probit", action = 'store_false', help = "Disable probit transformation", default = True)
    parser.add_option("--config", dest = "config", type = "string", help = "Config file. Warning: command line options override config options", default = None)
    
    (options, args) = parser.parse_args()

    if options.config is not None:
        options = load_options(options, parser)
    # Paths
    if options.input is not None:
        options.input = Path(options.input).resolve()
    elif options.config is not None:
        options.input = Path('.').resolve()
    else:
        raise Exception("Please provide path to samples.")
    if options.output is not None:
        options.output = Path(options.output).resolve()
        if not options.output.exists():
            options.output.mkdir(parents=True)
    else:
        options.output = options.input.parent
    if options.config is not None:
        options.config = Path(options.config).resolve()
    if options.config is None:
        save_options(options, options.output)

    # Read bounds
    if options.bounds is not None:
        options.bounds = np.array(np.atleast_2d(eval(options.bounds)), dtype = np.float64)
    elif options.bounds is None and not options.postprocess:
        raise Exception("Please provide bounds for the inference (use -b '[[xmin,xmax],[ymin,ymax],...]')")
    # Read parameter(s)
    if options.par is not None:
        options.par = options.par.split(',')
        if not np.all([par in supported_pars for par in options.par]):
            raise Exception("Please provide parameters from this list: "+', '.join(supported_pars[:-2]))
    # If provided, load injected density
    inj_density = None
    if options.inj_density_file is not None:
        inj_file_name = Path(options.inj_density_file).parts[-1].split('.')[0]
        spec = importlib.util.spec_from_file_location(inj_file_name, options.inj_density_file)
        inj_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(inj_module)
        inj_density = inj_module.density
    #If provided, load selecton function
    selfunc = None
    if options.selfunc_file is not None:
        selfunc, _, _, _ = load_selection_function(options.selfunc_file, par = options.par)
        if not callable(selfunc):
            raise Exception("Only .py files with callable approximants are allowed for DPGMM reconstruction")
    
    if options.sigma_prior is not None:
        options.sigma_prior = np.array([float(s) for s in options.sigma_prior.split(',')])
    if options.input.is_file():
        files = [options.input]
        output_draws = options.output
        subfolder = False
    else:
        files = sum([list(options.input.glob('*.'+ext)) for ext in supported_extensions], [])
        output_draws = Path(options.output, 'draws')
        if not output_draws.exists():
            output_draws.mkdir()
        subfolder = True
    if options.exclude_points:
        print("Ignoring points outside bounds.")
    
    for i, file in enumerate(files):
        # Load samples
        samples, name = load_single_event(file, par = options.par, n_samples = options.n_samples_dsp, cosmology = options.cosmology, waveform = options.wf, snr_threshold = options.snr_threshold, far_threshold = options.far_threshold)
        try:
            dim = np.shape(samples)[-1]
        except IndexError:
            dim = 1
        if options.exclude_points:
            samples = samples[np.where((np.prod(options.bounds[:,0] < samples, axis = 1) & np.prod(samples < options.bounds[:,1], axis = 1)))]
        else:
            # Check if all samples are within bounds
            if options.probit:
                if not np.all([(samples[:,i] > options.bounds[i,0]).all() and (samples[:,i] < options.bounds[i,1]).all() for i in range(dim)]):
                    raise ValueError("One or more samples are outside the given bounds.")

        # Reconstruction
        if not options.postprocess:
            # Actual analysis
            prior_pars = get_priors(options.bounds, samples = samples, std = options.sigma_prior, scale = options.fraction, probit = options.probit, hierarchical = False)
            mix        = DPGMM(options.bounds, prior_pars = prior_pars, probit = options.probit)
            desc       = name + ' ({0}/{1})'.format(i+1, len(files))
            draws      = np.array([mix.density_from_samples(samples, make_comp = False) for _ in tqdm(range(options.draws), desc = desc)])
            # Save reconstruction
            save_density(draws, folder = output_draws, name = 'draws_'+name, ext = options.ext)
        else:
            draws = load_density(Path(output_draws, 'draws_'+name+'.'+options.ext))

        # Plot
        if dim == 1:
            plot_median_cr(draws, injected = inj_density, selfunc = selfunc, samples = samples, out_folder = options.output, name = name, label = options.symbol, unit = options.unit, subfolder = subfolder)
        else:
            if options.symbol is not None:
                symbols = options.symbol.split(',')
            else:
                symbols = options.symbol
            if options.unit is not None:
                units = options.unit.split(',')
            else:
                units = options.unit
            plot_multidim(draws, samples = samples, out_folder = options.output, name = name, labels = symbols, units = units, subfolder = subfolder)

if __name__ == '__main__':
    main()



import numpy as np
import sys
import dill

from collections import Counter
from pathlib import Path

from scipy.special import gammaln, logsumexp
from scipy.stats import multivariate_normal as mn
from scipy.stats import invwishart, norm, invgamma, dirichlet, gamma

from figaro.decorators import *
from figaro.transform import *
from figaro._numba_functions import *
from figaro._likelihood import evaluate_mixture_MC_draws, evaluate_mixture_MC_draws_1d, log_norm
from figaro.exceptions import except_hook, FIGAROException
from figaro.utils import get_priors
from figaro.marginal import _condition, _marginalise

from numba import njit, prange

sys.excepthook = except_hook

#-----------#
# Functions #
#-----------#
@njit
def _student_t(df, t, mu, sigma, dim):
    """
    Multivariate student-t pdf.
    As in http://gregorygundersen.com/blog/2020/01/20/multivariate-t/
    
    Arguments:
        float df:         degrees of freedom
        float t:          variable (2d array)
        np.ndarray mu:    mean (2d array)
        np.ndarray sigma: variance
        int dim:          number of dimensions
        
    Returns:
        float: student_t(df).logpdf(t)
    """
    vals, vecs = eigh_jit(sigma)
    logdet     = np.log(vals).sum()
    U          = vecs * np.sqrt(1./vals)
    dev        = t - mu
    maha       = np.sum(dot_jit(dev, U)**2, axis=-1)

    x = 0.5 * (df + dim)
    A = gammaln_jit(x)
    B = gammaln_jit(0.5 * df)
    C = dim/2. * np.log(df * np.pi)
    D = 0.5 * logdet
    E = -x * np.log1p((1./df) * maha)

    return (A - B - C - D + E)[0]

@njit
def _update_alpha(alpha, n, K, alpha0, burnin = 1000):
    """
    Update concentration parameter using a Metropolis-Hastings sampling scheme.
    
    Arguments:
        double alpha: Initial value for concentration parameter
        int n:        Number of samples
        int K:        Number of active clusters
        int burnin:   MH burnin
    
    Returns:
        double: new concentration parameter value
    """
    a_old = alpha
    n_draws = burnin+np.random.randint(100)
    for i in prange(n_draws):
        a_new = a_old + (np.random.random() - 0.5)*0.5
        if a_new > 0.:
            logP_old = gammaln_jit(a_old) - gammaln_jit(a_old + n) + (K+alpha0) * np.log(a_old) - a_old
            logP_new = gammaln_jit(a_new) - gammaln_jit(a_new + n) + (K+alpha0) * np.log(a_new) - a_new
            if logP_new - logP_old > np.log(np.random.random()):
                a_old = a_new
    return a_old

@njit
def _compute_t_pars(k, mu, nu, L, mean, S, N, dim):
    """
    Compute parameters for student-t distribution.
    
    Arguments:
        double k:        Normal std parameter (for NIW)
        np.ndarray mu:   Normal mean parameter (for NIW)
        int nu:          Inverse-Wishart df parameter (for NIW)
        np.ndarray L:    Inverse-Wishart scale matrix (for NIW)
        np.ndarray mean: samples mean
        np.ndarray S:    samples covariance
        int N:           number of samples
        int dim:         number of dimensions
    
    Returns:
        int:        degrees of fredom for student-t
        np.ndarray: scale matrix for student-t
        np.ndarray: mean for student-t
    """
    # Update hyperparameters
    k_n, mu_n, nu_n, L_n = _compute_hyperpars(k, mu, nu, L, mean, S, N)
    # Update t-parameters
    t_df    = nu_n - dim + 1
    t_shape = rescale_matrix(L_n, 1./((k_n+1.)/(k_n*t_df)))
    return t_df, t_shape, mu_n

@njit
def _compute_hyperpars(k, mu, nu, L, mean, S, N):
    """
    Update hyperparameters for Normal Inverse Gamma/Wishart (NIG/NIW).
    See https://www.cs.ubc.ca/~murphyk/Papers/bayesGauss.pdf
    
    Arguments:
        double k:        Normal std parameter (for NIG/NIW)
        np.ndarray mu:   Normal mean parameter (for NIG/NIW)
        int nu:          Gamma df parameter (for NIG/NIW)
        np.ndarray L:    Gamma scale matrix (for NIG/NIW)
        np.ndarray mean: samples mean
        np.ndarray S:    samples covariance
        int N:           number of samples
    
    Returns:
        double:     updated Normal std parameter (for NIG/NIW)
        np.ndarray: updated Normal mean parameter (for NIG/NIW)
        int:        updated Gamma df parameter (for NIG/NIW)
        np.ndarray: updated Gamma scale matrix (for NIG/NIW)
    """
    k_n  = k + N
    mu_n = (mu*k + N*mean)/k_n
    nu_n = nu + N
    if N > 0:
        L_n = L + S + rescale_matrix(outer_jit((mean - mu).T, (mean - mu)), k_n/(k*N))
    else:
        L_n = L + S
    return k_n, mu_n, nu_n, L_n

@njit
def _compute_component_suffstats_add(x, mean, S, N, p_mu, p_k, p_nu, p_L):
    """
    Update mean, covariance, number of samples and maximum a posteriori for mean and covariance.
    
    Arguments:
        np.ndarray x:    sample to add
        np.ndarray mean: mean of samples already in the cluster
        np.ndarray cov:  covariance of samples already in the cluster
        int N:           number of samples already in the cluster
        np.ndarray p_mu: NIG Normal mean parameter
        double p_k:      NIG Normal std parameter
        int p_nu:        NIG Gamma df parameter
        np.ndarray p_L:  NIG Gamma scale matrix
    
    Returns:
        np.ndarray: updated mean
        np.ndarray: updated covariance
        int:        updated number of samples
        np.ndarray: mean (maximum a posteriori)
        np.ndarray: covariance (maximum a posteriori)
    """
    new_mean  = (mean*N+x)/(N+1)
    new_S     = (S + rescale_matrix(outer_jit(mean,mean), 1./N) + outer_jit(x,x)) - rescale_matrix(outer_jit(new_mean, new_mean), 1./(N+1))
    new_N     = N+1
    new_mu    = ((p_mu*p_k + new_N*new_mean)/(p_k + new_N))[0]
    new_sigma = rescale_matrix(p_L + new_S + rescale_matrix(outer_jit((new_mean - p_mu).T, (new_mean - p_mu)), (p_k + new_N)/(p_k*new_N)), (p_nu + new_N - x.shape[-1] - 1))
    
    return new_mean, new_S, new_N, new_mu, new_sigma

@njit
def _compute_component_suffstats_remove(x, mean, S, N, p_mu, p_k, p_nu, p_L):
    """
    Update mean, covariance, number of samples and maximum a posteriori for mean and covariance.
    
    Arguments:
        np.ndarray x:    sample to remove
        np.ndarray mean: mean of samples already in the cluster
        np.ndarray cov:  covariance of samples already in the cluster
        int N:           number of samples already in the cluster
        np.ndarray p_mu: NIG Normal mean parameter
        double p_k:      NIG Normal std parameter
        int p_nu:        NIG Gamma df parameter
        np.ndarray p_L:  NIG Gamma scale matrix
    
    Returns:
        np.ndarray: updated mean
        np.ndarray: updated covariance
        int:        updated number of samples
        np.ndarray: mean (maximum a posteriori)
        np.ndarray: covariance (maximum a posteriori)
    """
    if N > 1:
        new_mean  = (mean*N-x)/(N-1)
        if N > 2:
            new_S = (S + rescale_matrix(outer_jit(mean,mean), 1./N) - outer_jit(x,x)) - rescale_matrix(outer_jit(new_mean, new_mean), 1./(N-1))
        else:
            new_S = np.zeros(S.shape)
    else:
        new_mean  = np.zeros(mean.shape)
        new_S     = np.zeros(S.shape)
    new_N     = N-1
    new_mu    = ((p_mu*p_k + new_N*new_mean)/(p_k + new_N))[0]
    if N > 2:
        new_sigma = rescale_matrix(p_L + new_S + rescale_matrix(outer_jit((new_mean - p_mu).T, (new_mean - p_mu)), (p_k + new_N)/(p_k*new_N)), (p_nu + new_N - x.shape[-1] - 1))
    else:
        new_sigma = rescale_matrix(p_L, (p_nu + new_N - x.shape[-1] - 1))
    return new_mean, new_S, new_N, new_mu, new_sigma
    
#-------------------#
# Auxiliary classes #
#-------------------#

class _prior:
    """
    Class to store the NIW prior parameters
    See https://www.cs.ubc.ca/~murphyk/Papers/bayesGauss.pdf, sec. 9
    
    Arguments:
        double k:        Normal std parameter
        np.ndarray L:    Wishart scale matrix
        int nu:          Wishart df parameter
        np.ndarray mu:   Normal mean parameter
    
    Returns:
        prior: instance of prior class
    """
    def __init__(self, k, L, nu, mu):
         self.k   = k
         self.nu  = np.max([nu, mu.shape[-1]+2])
         self.L   = rescale_matrix(L, 1./(self.nu-mu.shape[-1]-1))
         self.mu  = mu

class _component:
    """
    Class to store the relevant informations for each component in the mixture.
    
    Arguments:
        np.ndarray x: sample added to the new component
        prior prior:  instance of the prior class with NIG/NIW prior parameters
    
    Returns:
        component: instance of component class
    """
    def __init__(self, x, prior, N = 1):
        self.N     = N
        self.mean  = x
        self.S     = np.identity(x.shape[-1])*0.
        self.mu    = np.atleast_2d((prior.mu*prior.k + self.N*self.mean)/(prior.k + self.N)).astype(np.float64)[0]
        self.sigma = rescale_matrix(prior.L, (prior.nu - x.shape[-1] - 1))

class _component_h:
    """
    Class to store the relevant informations for each component in the mixture.
    To be used in hierarchical inference.
    
    Arguments:
        np.ndarray x:  event added to the new component
        int dim:       number of dimensions
        prior prior:   instance of the prior class with NIG/NIW prior parameters
        double logL_D: logLikelihood denominator
    
    Returns:
        component_h: instance of component_h class
    """
    def __init__(self, x, dim, prior, logL_D, mu_MC, sigma_MC, log_alpha_factor):
        self.dim    = dim
        self.N      = 1
        self.logL_D = logL_D
        
        log_norm_D = logsumexp_jit(logL_D)
        
        idx        = np.random.choice(len(mu_MC), p = np.exp(logL_D - log_norm_D))
        self.mu    = np.copy(mu_MC[idx])
        self.sigma = np.copy(sigma_MC[idx])
        if dim == 1:
            self.mu = np.atleast_2d(self.mu).T
            self.sigma = np.atleast_2d(self.sigma).T
        self.log_N_true = np.log(self.N) - log_alpha_factor[idx]
            
class density:
    """
    Class to initialise a common set of methods for mixture models. Not to be used.
    """
    def __init__(self):
        pass
        
    def __call__(self, x):
        return self.pdf(x)

    def pdf(self, x):
        if self.n_cl == 0:
            raise FIGAROException("You are trying to evaluate an empty mixture.\n If you are using the density_from_samples() method, you may want to evaluate the output of that method.")
        if len(np.shape(x)) < 2:
            if self.dim == 1:
                x = np.atleast_2d(x).T
            else:
                x = np.atleast_2d(x)
        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            p = np.nan_to_num(self._pdf(x), nan = 0.)
        return p

    def logpdf(self, x):
        if self.n_cl == 0:
            raise FIGAROException("You are trying to evaluate an empty mixture.\n If you are using the density_from_samples() method, you may want to evaluate the output of that method.")
        if len(np.shape(x)) < 2:
            if self.dim == 1:
                x = np.atleast_2d(x).T
            else:
                x = np.atleast_2d(x)
        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            logp = np.nan_to_num(self._logpdf(x), nan = -np.inf)
        return logp

    @probit
    def _pdf(self, x):
        """
        Evaluate mixture at point(s) x
        
        Arguments:
            np.ndarray x: point(s) to evaluate the mixture at
        
        Returns:
            np.ndarray: mixture.pdf(x)
        """
        return self._pdf_probit(x) * np.exp(-probit_logJ(x, self.bounds, self.probit))

    @probit
    def _logpdf(self, x):
        """
        Evaluate log mixture at point(s) x
        
        Arguments:
            np.ndarray x: point(s) to evaluate the mixture at
        
        Returns:
            np.ndarray: mixture.logpdf(x)
        """
        return self._logpdf_probit(x) - probit_logJ(x, self.bounds, self.probit)

    def fast_pdf(self, x):
        """
        Fast pdf evaluation using FIGARO implementation of log_norm (JIT) rather than Numpy's.
        WARNING: it is meant to be used with MCMC samplers, therefore accepts only one point at a time.
        
        Arguments:
            np.ndarray x: point to evaluate the mixture at
        
        Returns:
            np.ndarray: mixture.pdf(x)
        """
        x = np.atleast_1d(x)
        if x.shape == (1,self.dim):
            return self._fast_pdf(x[0])
        elif x.shape == (self.dim,):
            return self._fast_pdf(x)
        else:
            raise FIGAROException("Please provide one point at a time.")

    def fast_logpdf(self, x):
        """
        Fast logpdf evaluation using FIGARO implementation of log_norm (JIT) rather than Numpy's.
        WARNING: it is meant to be used with MCMC samplers, therefore accepts only one point at a time.
        
        Arguments:
            np.ndarray x: point to evaluate the mixture at
        
        Returns:
            np.ndarray: mixture.pdf(x)
        """
        x = np.atleast_1d(x)
        if x.shape == (1,self.dim):
            return self._fast_logpdf(x[0])
        elif x.shape == (self.dim,):
            return self._fast_logpdf(x)
        else:
            raise FIGAROException("Please provide one point at a time.")

    @probit
    def _fast_pdf(self, x):
        """
        Evaluate mixture at point x
        
        Arguments:
            np.ndarray x: point to evaluate the mixture at
        
        Returns:
            np.ndarray: mixture.pdf(x)
        """
        return self._fast_pdf_probit(x) * np.exp(-probit_logJ(x, self.bounds, self.probit))

    @probit
    def _fast_logpdf(self, x):
        """
        Evaluate log mixture at point x
        
        Arguments:
            np.ndarray x: point to evaluate the mixture at
        
        Returns:
            np.ndarray: mixture.logpdf(x)
        """
        return self._fast_logpdf_probit(x) - probit_logJ(x, self.bounds, self.probit)

    def _fast_pdf_probit(self, x):
        """
        Evaluate mixture at point x in probit space
        
        Arguments:
            np.ndarray x: point to evaluate the mixture at (in probit space)
        
        Returns:
            np.ndarray: mixture.pdf(x)
        """
        return np.sum(np.array([w*np.exp(log_norm(x[0], mean, cov)) for mean, cov, w in zip(self.means, self.covs, self.w)]), axis = 0)

    def _fast_logpdf_probit(self, x):
        """
        Evaluate log mixture at point x in probit space
        
        Arguments:
            np.ndarray x: point to evaluate the mixture at (in probit space)
        
        Returns:
            np.ndarray: mixture.logpdf(x)
        """
        return logsumexp(np.array([w + log_norm(x[0], mean, cov) for mean, cov, w in zip(self.means, self.covs, self.log_w)]), axis = 0)

    @probit
    def _pdf_no_jacobian(self, x):
        """
        Evaluate mixture at point(s) x without jacobian
        
        Arguments:
            np.ndarray x: point(s) to evaluate the mixture at
        
        Returns:
            np.ndarray: mixture.pdf(x)
        """
        return self._pdf_probit(x)

    def _pdf_probit(self, x):
        """
        Evaluate mixture at point(s) x in probit space
        
        Arguments:
            np.ndarray x: point(s) to evaluate the mixture at (in probit space)
        
        Returns:
            np.ndarray: mixture.pdf(x)
        """
        return np.sum(np.array([w*mn(mean, cov, allow_singular = True).pdf(x) for mean, cov, w in zip(self.means, self.covs, self.w)]), axis = 0)
    
    @probit
    def _pdf_array(self, x):
        """
        Evaluate every mixture component at point(s) x.
        
        Arguments:
            np.ndarray x: point(s) to evaluate the components at
        
        Returns:
            np.ndarray: component.pdf(x) for each mixture component
        """
        return _pdf_array_probit(x) * np.exp(-probit_logJ(x, self.bounds, self.probit))

    def _pdf_array_probit(self, x):
        """
        Evaluate every mixture component at point(s) x.
        
        Arguments:
            np.ndarray x: point(s) to evaluate the components at (in probit space)
        
        Returns:
            np.ndarray: component.pdf(x) for each mixture component
        """
        return np.array([w*mn(mean, cov, allow_singular = True).pdf(x) for mean, cov, w in zip(self.means, self.covs, self.w)])

    @probit
    def _fast_pdf_array(self, x):
        """
        Evaluate every mixture component at point(s) x.
        
        Arguments:
            np.ndarray x: point(s) to evaluate the components at
        
        Returns:
            np.ndarray: component.pdf(x) for each mixture component
        """
        return _fast_pdf_array_probit(x) * np.exp(-probit_logJ(x, self.bounds, self.probit))

    def _fast_pdf_array_probit(self, x):
        """
        Evaluate every mixture component at point(s) x.
        
        Arguments:
            np.ndarray x: point(s) to evaluate the components at (in probit space)
        
        Returns:
            np.ndarray: component.pdf(x) for each mixture component
        """
        return np.array([w*np.exp(log_norm(x[0], mean, cov)) for mean, cov, w in zip(self.means, self.covs, self.w)])

    @probit
    def _logpdf_no_jacobian(self, x):
        """
        Evaluate log mixture at point(s) x without jacobian
        
        Arguments:
            np.ndarray x: point(s) to evaluate the mixture at
        
        Returns:
            np.ndarray: mixture.logpdf(x)
        """
        return self._logpdf_probit(x)

    def _logpdf_probit(self, x):
        """
        Evaluate log mixture at point(s) x in probit space
        
        Arguments:
            np.ndarray x: point(s) to evaluate the mixture at (in probit space)
        
        Returns:
            np.ndarray: mixture.logpdf(x)
        """
        return logsumexp(np.array([w + mn(mean, cov, allow_singular = True).logpdf(x) for mean, cov, w in zip(self.means, self.covs, self.log_w)]), axis = 0)

    def cdf(self, x):
        if self.dim > 1:
            raise FIGAROException("cdf is provided only for 1-dimensional distributions")
        if len(np.shape(x)) < 2:
            x = np.atleast_2d(x).T
        return self._cdf(x)

    def logcdf(self, x):
        if self.dim > 1:
            raise FIGAROException("cdf is provided only for 1-dimensional distributions")
        if len(np.shape(x)) < 2:
            x = np.atleast_2d(x).T
        return self._logcdf(x)

    @probit
    def _cdf(self, x):
        """
        Evaluate mixture cdf at point(s) x
        
        Arguments:
            np.ndarray x: point(s) to evaluate the mixture at
        
        Returns:
            np.ndarray: mixture.cdf(x)
        """
        return np.sum(np.array([w*norm(mean[0], cov[0,0]).cdf(x) for mean, cov, w in zip(self.means, np.sqrt(self.covs), self.w)]), axis = 0)

    @probit
    def _logcdf(self, x):
        """
        Evaluate mixture log cdf at point(s) x
        
        Arguments:
            np.ndarray x: point(s) to evaluate the mixture at
        
        Returns:
            np.ndarray: mixture.logcdf(x)
        """
        return logsumexp(np.array([w + norm(mean[0], cov[0,0]).logcdf(x) for mean, cov, w in zip(self.means, np.sqrt(self.covs), self.log_w)]), axis = 0)

    @from_probit
    def rvs(self, size = 1):
        """
        Draw samples from mixture
        
        Arguments:
            int size: number of samples to draw
        
        Returns:
            np.ndarray: samples
        """
        if self.n_cl == 0:
            raise FIGAROException("You are trying to draw samples from an empty mixture.\n If you are using the density_from_samples() method, you may want to draw samples from the output of that method.")
        return self._rvs_probit(size)
        
    def _rvs_probit(self, size = 1):
        """
        Draw samples from mixture in probit space
        
        Arguments:
            int size: number of samples to draw
        
        Returns:
            np.ndarray: samples in probit space
        """
        idx = np.random.choice(np.arange(self.n_cl), p = self.w, size = size)
        ctr = Counter(idx)
        if self.dim > 1:
            samples = np.empty(shape = (1,self.dim))
            for i, n in zip(ctr.keys(), ctr.values()):
                samples = np.concatenate((samples, np.atleast_2d(mn(self.means[i], self.covs[i], allow_singular = True).rvs(size = n))))
        else:
            samples = np.array([np.zeros(1)])
            for i, n in zip(ctr.keys(), ctr.values()):
                samples = np.concatenate((samples, np.atleast_2d(mn(self.means[i], self.covs[i], allow_singular = True).rvs(size = n)).T))
        return np.array(samples[1:])
    
    def gradient(self, x):
        """
        Gradient of the mixture.
        
        Arguments:
            np.ndarray x: point to evaluate the gradient at
        
        Returns:
            np.ndarray: gradient
        """
        if self.n_cl == 0:
            raise FIGAROException("You are trying to evaluate an empty mixture.\n If you are using the density_from_samples() method, you may want to evaluate the output of that method.")
        if len(np.shape(x)) < 2:
            if self.dim == 1:
                x = np.atleast_2d(x).T
            else:
                x = np.atleast_2d(x)
        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            g = np.nan_to_num([self._gradient(xi) for xi in x], nan = 0.)
        return g
    
    def log_gradient(self, x):
        """
        Logarithmic gradient of the mixture.
        
        Arguments:
            np.ndarray x: point to evaluate the gradient at
        
        Returns:
            np.ndarray: logarithmic gradient
        """
        if self.n_cl == 0:
            raise FIGAROException("You are trying to evaluate an empty mixture.\n If you are using the density_from_samples() method, you may want to evaluate the output of that method.")
        if len(np.shape(x)) < 2:
            if self.dim == 1:
                x = np.atleast_2d(x).T
            else:
                x = np.atleast_2d(x)
        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            g = np.nan_to_num([self._log_gradient(xi) for xi in x], nan = 0.)
        return g
    
    def _gradient(self, x):
        """
        Gradient of the mixture.
        
        Arguments:
            np.ndarray x: point to evaluate the gradient at
        
        Returns:
            np.ndarray: gradient
        """
        return self._pdf(x)*self._log_gradient(x)
    
    @probit
    def _log_gradient(self, x):
        """
        Logarithmic gradient of the mixture.
        
        Arguments:
            np.ndarray x: point to evaluate the gradient at
        
        Returns:
            np.ndarray: logarithmic gradient
        """
        p = self._pdf_array_probit(x)
        J = np.exp(-probit_log_jacobian(x, self.bounds, self.probit))
        B = np.array([-dot_jit(inv_jit(sigma),(x - mu))*J for mu, sigma in zip(self.means, self.covs)])
        try:
            return np.average(B, weights = p, axis = 0) + gradient_inv_jacobian(x, self.bounds, self.probit)*J
        except ZeroDivisionError:
            return np.zeros(x.shape[-1])

class mixture(density):
    """
    Class to store a single draw from DPGMM/(H)DPGMM.
    Methods inherited from density class
    
    Arguments:
        iterable means:      component means
        iterable covs:       component covariances
        np.ndarray w:        component weights
        np.ndarray bounds:   bounds of probit transformation
        int dim:             number of dimensions
        int n_cl:            number of clusters in the mixture
        int n_pts:           number of points used to infer the mixture
        double alpha:        concentration parameter
        bool probit:         whether to use the probit transformation or not
        np.ndarray log_w:    component log weights
        bool make_comp:      make component objects
        double alpha_factor: evaluated \int pdet(theta)p(theta|lambda) conditioned on the mixture parameters
    
    Returns:
        mixture: instance of mixture class
    """
    def __init__(self, means, covs, w, bounds, dim, n_cl, n_pts, alpha = 1., probit = True, log_w = None, make_comp = True, alpha_factor = 1.):
        self.means = means
        self.covs  = covs
        if make_comp:
            self.components = [mn(mean, cov, allow_singular = True) for mean, cov in zip(self.means, self.covs)]
        else:
            self.components = None
        if log_w is None:
            self.w      = w
            self.log_w  = np.log(w)
        else:
            self.log_w  = log_w
            self.w        = np.exp(log_w)
        self.bounds       = bounds
        self.dim          = int(dim)
        self.n_cl         = int(n_cl)
        self.n_pts        = int(n_pts)
        self.probit       = probit
        self.alpha        = alpha
        self.alpha_factor = alpha_factor
    
    def marginalise(self, axis = -1):
        """
        Marginalise out one or more dimensions from the mixture.
        
        Arguments:
            int or list of int axis: axis to marginalise on. Default: last
        
        Returns:
            figaro.mixture.mixture: marginalised mixture
        """
        return _marginalise(self, axis)
    
    def condition(self, vals, dims, norm = True, filter = True, tol = 1e-3):
        """
        Mixture conditioned on specific values of a subset of parameters.
        
        Arguments:
            iterable vals:           value(s) to condition on
            int or list of int dims: dimension(s) associated with given vals (starting from 0)
            bool norm:               whether to normalize the distribution or not
            bool filter:             filter the components with weight < tol
            double tol:              tolerance on the sum of the weights
        
        Returns:
            figaro.mixture.mixture: conditioned mixture
        """
        v       = np.mean(self.bounds, axis = -1)
        v[dims] = vals
        return _condition(self, v, dims, norm, filter, tol = 1e-3)

    def _logpdf_probit(self, x):
        """
        Evaluate log mixture at point(s) x in probit space
        
        Arguments:
            np.ndarray x: point(s) to evaluate the mixture at (in probit space)
        
        Returns:
            np.ndarray: mixture.logpdf(x)
        """
        if self.components is not None:
            return logsumexp(np.array([w+comp.logpdf(x) for w, comp in zip(self.log_w, self.components)]), axis = 0)
        else:
            return super()._logpdf_probit(x)

    def _pdf_probit(self, x):
        """
        Evaluate mixture at point(s) x in probit space
        
        Arguments:
            np.ndarray x: point(s) to evaluate the mixture at (in probit space)
        
        Returns:
            np.ndarray: mixture.pdf(x)
        """
        if self.components is not None:
            return np.sum(np.array([w*comp.pdf(x) for w, comp in zip(self.w, self.components)]), axis = 0)
        else:
            return super()._pdf_probit(x)

    def _pdf_array_probit(self, x):
        """
        Evaluate every mixture component at point(s) x.
        
        Arguments:
            np.ndarray x: point(s) to evaluate the components at (in probit space)
        
        Returns:
            np.ndarray: component.pdf(x) for each mixture component
        """
        if self.components is not None:
            return np.array([w*comp.pdf(x) for w, comp in zip(self.w, self.components)])
        else:
            return super()._pdf_array_probit(x)

#-------------------#
# Inference classes #
#-------------------#

class DPGMM(density):
    """
    Class to infer a distribution given a set of samples.
    
    Arguments:
        iterable bounds:     boundaries of the rectangle over which the distribution is defined. It should be in the format [[xmin, xmax],[ymin, ymax],...]
        iterable prior_pars: NIW prior parameters (k, L, nu, mu)
        double alpha0:       initial guess for concentration parameter
        bool probit:         whether to use the probit transformation or not
        int n_reassignments: number of reassignments. Default is not to reassign.
    
    Returns:
        DPGMM: instance of DPGMM class
    """
    def __init__(self, bounds,
                       prior_pars      = None,
                       alpha0          = 1.,
                       probit          = True,
                       n_reassignments = 0.,
                       ):
        self.probit = probit
        self.bounds = np.atleast_2d(bounds)
        self.dim    = len(self.bounds)
        if prior_pars is not None:
            self.prior = _prior(*prior_pars)
        else:
            self.prior = _prior(*get_priors(bounds = self.bounds, probit = self.probit))
        self.alpha           = alpha0
        self.alpha_0         = alpha0
        self.mixture         = []
        self.w               = []
        self.log_w           = []
        self.N_list          = []
        self.n_cl            = 0
        self.n_pts           = 0
        self.stored_pts      = {}
        self.assignations    = {}
        self.n_reassignments = int(n_reassignments)

    def __call__(self, x):
        return self.pdf(x)

    def initialise(self, prior_pars = None):
        """
        Initialise the mixture to initial conditions.
        
        Arguments:
            iterable prior_pars: NIW prior parameters (k, L, nu, mu). If None, old parameters are kept
        """
        self.alpha        = self.alpha_0
        self.mixture      = []
        self.w            = []
        self.log_w        = []
        self.N_list       = []
        self.n_cl         = 0
        self.n_pts        = 0
        self.stored_pts   = {}
        self.assignations = {}
        if prior_pars is not None:
            self.prior = _prior(*prior_pars)
        
    def _add_datapoint_to_component(self, x, ss):
        """
        Update component parameters after assigning a sample to a component
        
        Arguments:
            np.ndarray x: sample
            component ss: component to update
        
        Returns:
            component: updated component
        """
        new_mean, new_S, new_N, new_mu, new_sigma = _compute_component_suffstats_add(x, ss.mean, ss.S, ss.N, self.prior.mu, self.prior.k, self.prior.nu, self.prior.L)
        ss.mean  = new_mean
        ss.S     = new_S
        ss.N     = new_N
        ss.mu    = new_mu
        ss.sigma = new_sigma
        return ss

    def _remove_datapoint_from_component(self, x, ss):
        """
        Update component parameters after removing a sample from the component
        
        Arguments:
            np.ndarray x: sample
            component ss: component to update
        
        Returns:
            component: updated component
        """
        new_mean, new_S, new_N, new_mu, new_sigma = _compute_component_suffstats_remove(x, ss.mean, ss.S, ss.N, self.prior.mu, self.prior.k, self.prior.nu, self.prior.L)
        ss.mean  = new_mean
        ss.S     = new_S
        ss.N     = new_N
        ss.mu    = new_mu
        ss.sigma = new_sigma
        return ss

    def _log_predictive_likelihood(self, x, ss):
        """
        Compute log likelihood of drawing sample x from component ss given the samples that are already assigned to that component.
        
        Arguments:
            np.ndarray x: sample
            component ss: component to update
        
        Returns:
            double: log Likelihood
        """
        if ss is None:
            ss = _component(self.prior.mu, prior = self.prior, N = 0)
        t_df, t_shape, mu_n = _compute_t_pars(self.prior.k, self.prior.mu, self.prior.nu, self.prior.L, ss.mean, ss.S, ss.N, self.dim)
        try:
            return _student_t(df = t_df, t = x, mu = mu_n, sigma = t_shape, dim = self.dim)
        except np.linalg.LinAlgError:
            return -np.inf

    def _cluster_assignment_distribution(self, x):
        """
        Compute the marginal distribution of cluster assignment for each cluster.
        
        Arguments:
            np.ndarray x: sample
        
        Returns:
            dict: p_i for each component
        """
        scores = np.zeros(self.n_cl+1)
        for i in range(self.n_cl+1):
            if i == 0:
                ss        = None
                scores[i] = np.log(self.alpha)
            else:
                ss        = self.mixture[i-1]
                with np.errstate(divide = 'ignore', invalid = 'ignore'):
                    scores[i] = np.log(ss.N)
            if np.isfinite(scores[i]):
                scores[i] += self._log_predictive_likelihood(x, ss)
            else:
                scores[i] = -np.inf
        norm = logsumexp_jit(scores)
        return np.exp(scores - norm)

    def _assign_to_cluster(self, x, pt_id = None):
        """
        Assign the new sample x to an existing cluster or to a new cluster according to the marginal distribution of cluster assignment.
        
        Arguments:
            np.ndarray x: sample
            int pt_id:    point id
        """
        if pt_id is None:
            pt_id = len(list(self.stored_pts.keys()))-1
        self.n_pts += 1
        scores = self._cluster_assignment_distribution(x)
        cid = np.random.choice(np.arange(-1,self.n_cl), p=scores)
        if cid == -1:
            self.mixture.append(_component(x, prior = self.prior))
            self.N_list.append(1.)
            self.n_cl += 1
            self.assignations[pt_id] = int(self.n_cl)-1
        else:
            self.mixture[int(cid)] = self._add_datapoint_to_component(x, self.mixture[int(cid)])
            self.N_list[int(cid)] += 1
            self.assignations[pt_id] = int(cid)
        # Update weights
        self.w = np.array(self.N_list)
        # Beware of empty clusters!
        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            self.w     = self.w/self.w.sum()
            self.log_w = np.log(self.w)
        return
    
    def _remove_from_cluster(self, x, cid):
        """
        Remove the sample x from its assigned cluster.
        
        Arguments:
            np.ndarray x: sample
            int cid:      cluster id
        """
        self.mixture[int(cid)] = self._remove_datapoint_from_component(x, self.mixture[int(cid)])
        self.N_list[int(cid)] -= 1
        self.n_pts            -= 1
        # Update weights
        self.w = np.array(self.N_list)
        # Beware of empty clusters!
        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            self.w = self.w/self.w.sum()
            self.log_w = np.log(self.w)
        return
    
    def density_from_samples(self, samples, make_comp = True):
        """
        Reconstruct the probability density from a set of samples.
        
        Arguments:
            iterable samples: samples set
            bool make_comp:   whether to instantiate the scipy.stats.multivariate_normal components or not
        
        Returns:
            mixture: the inferred mixture
        """
        np.random.shuffle(samples)
        samples = np.ascontiguousarray(samples)
        # Thermalise
        for s in samples:
            self.add_new_point(s)
        # Random Gibbs walk (if required)
        for id in np.random.choice(self.n_pts, size = self.n_reassignments, replace = True):
            self._reassign_point(id)
        d = self.build_mixture(make_comp)
        self.initialise()
        return d
    
    @probit
    def add_new_point(self, x):
        """
        Update the probability density reconstruction adding a new sample
        
        Arguments:
            np.ndarray x: sample
        """
        self.stored_pts[len(list(self.stored_pts.keys()))] = np.atleast_2d(x)
        self._assign_to_cluster(np.atleast_2d(x))
        self.alpha = _update_alpha(self.alpha, self.n_pts, self.n_cl, self.alpha_0)
    
    def _reassign_point(self, id):
        """
        Update the probability density reconstruction reassigining an existing sample
        
        Arguments:
            id: sample id
        """
        x   = self.stored_pts[id]
        cid = self.assignations[id]
        self._remove_from_cluster(x, cid)
        self._assign_to_cluster(x, id)
        self.alpha = _update_alpha(self.alpha, self.n_pts, (np.array(self.N_list) > 0).sum(), self.alpha_0)

    def build_mixture(self, make_comp = True):
        """
        Instances a mixture class representing the inferred distribution
        
        Arguments:
            bool make_comp:   whether to instantiate the scipy.stats.multivariate_normal components or not
        
        Returns:
            mixture: the inferred distribution
        """
        if self.n_cl == 0:
            raise FIGAROException("You are trying to build an empty mixture - perhaps you called the initialise() method. If you are using the density_from_samples() method, the inferred mixture is returned by that method as an instance of mixture class.")
        # Number of active clusters
        n_cl      = (np.array(self.N_list) > 0).sum()
        means     = np.zeros((n_cl, self.dim))
        variances = np.zeros((n_cl, self.dim, self.dim))
        i = 0
        for ss in self.mixture:
            if ss.N > 0:
                k_n, mu_n, nu_n, L_n = _compute_hyperpars(self.prior.k, self.prior.mu, self.prior.nu, self.prior.L, ss.mean, ss.S, ss.N)
                variances[i] = invwishart(df = nu_n, scale = L_n).rvs()
                means[i]     = mn(mean = mu_n[0], cov = rescale_matrix(variances[i], k_n), allow_singular = True).rvs()
                i += 1
        w = dirichlet(self.w[self.w > 0]*self.n_pts+self.alpha/self.n_cl).rvs()[0]
        return mixture(means, variances, w, self.bounds, self.dim, n_cl, self.n_pts, self.alpha, probit = self.probit, make_comp = make_comp)

    # Methods to overwrite density methods
    def _rvs_probit(self, size = 1):
        """
        Draw samples from mixture in probit space
        
        Arguments:
            int size: number of samples to draw
        
        Returns:
            np.ndarray: samples in probit space
        """
        idx = np.random.choice(np.arange(self.n_cl), p = self.w, size = size)
        ctr = Counter(idx)
        if self.dim > 1:
            samples = np.empty(shape = (1,self.dim))
            for i, n in zip(ctr.keys(), ctr.values()):
                samples = np.concatenate((samples, np.atleast_2d(mn(self.mixture[i].mu, self.mixture[i].sigma, allow_singular = True).rvs(size = n))))
        else:
            samples = np.array([np.zeros(1)])
            for i, n in zip(ctr.keys(), ctr.values()):
                samples = np.concatenate((samples, np.atleast_2d(mn(self.mixture[i].mu[0], self.mixture[i].sigma, allow_singular = True).rvs(size = n)).T))
        return samples[1:]

    def _pdf_probit(self, x):
        """
        Evaluate mixture at point(s) x in probit space
        
        Arguments:
            np.ndarray x: point(s) to evaluate the mixture at (in probit space)
        
        Returns:
            np.ndarray: mixture.pdf(x)
        """
        return np.sum(np.array([w*mn(comp.mu, comp.sigma, allow_singular = True).pdf(x) for comp, w in zip(self.mixture, self.w)]), axis = 0)

    def _logpdf_probit(self, x):
        """
        Evaluate log mixture at point(s) x in probit space
        
        Arguments:
            np.ndarray x: point(s) to evaluate the mixture at (in probit space)
        
        Returns:
            np.ndarray: mixture.logpdf(x)
        """
        return logsumexp(np.array([w + mn(comp.mu, comp.sigma, allow_singular = True).logpdf(x) for comp, w in zip(self.mixture, self.log_w)]), axis = 0)

    def _fast_pdf_probit(self, x):
        """
        Evaluate mixture at point x in probit space
        
        Arguments:
            np.ndarray x: point to evaluate the mixture at (in probit space)
        
        Returns:
            np.ndarray: mixture.pdf(x)
        """
        return np.sum(np.array([w*np.exp(log_norm(x[0], comp.mean, comp.cov)) for comp, w in zip(self.mixture, self.w)]), axis = 0)

    def _fast_logpdf_probit(self, x):
        """
        Evaluate log mixture at point x in probit space
        
        Arguments:
            np.ndarray x: point to evaluate the mixture at (in probit space)
        
        Returns:
            np.ndarray: mixture.logpdf(x)
        """
        return logsumexp(np.array([w + log_norm(x[0], comp.mean, comp.cov) for comp, w in zip(self.mixture, self.log_w)]), axis = 0)

class HDPGMM(DPGMM):
    """
    Class to infer a distribution given a set of observations (each being a set of samples).
    Child of DPGMM class
    
    Arguments:
        iterable bounds:             boundaries of the rectangle over which the distribution is defined. It should be in the format [[xmin, xmax],[ymin, ymax],...]
        iterable prior_pars:         IW parameters
        double alpha0:               initial guess for concentration parameter
        double MC_draws:             number of MC draws for integral
        bool probit:                 whether to use the probit transformation or not
        int n_reassignments:         number of reassignments. Default is not to reassign.
        callable selection_function: selection function approximant or samples
        np.ndarray injection_pdf:    pdf of injected samples (for selection function inclusion)
    
    Returns:
        HDPGMM: instance of HDPGMM class
    """
    def __init__(self, bounds,
                       alpha0             = 1.,
                       prior_pars         = None,
                       MC_draws           = None,
                       probit             = True,
                       n_reassignments    = 0.,
                       selection_function = None,
                       injection_pdf      = None,
                       total_injections   = None,
                       ):
        bounds   = np.atleast_2d(bounds)
        self.dim = len(bounds)
        super().__init__(bounds = bounds, alpha0 = alpha0, probit = probit, n_reassignments = n_reassignments)
        if prior_pars is not None:
            self.exp_sigma, self.a = prior_pars
        else:
            self.exp_sigma, self.a = get_priors(bounds = self.bounds, probit = self.probit, hierarchical = True)
        self.invgamma = invgamma(self.a)
        if MC_draws is None:
            self.MC_draws = int((self.dim+1)*1e3)
        else:
            self.MC_draws = int(MC_draws)
        self.evaluated_logL = {}
        # Selection function
        self.selfunc = selection_function
        if not callable(self.selfunc) and self.selfunc is not None:
            try:
                self.log_inj_pdf = np.log(injection_pdf)
                self.total_inj   = int(total_injections)
            except TypeError:
                raise FIGAROException("Please provide injection pdf")
            self.selfunc_probit   = self.selfunc
            self.log_jacobian_inj = np.zeros(len(self.selfunc))
            if self.probit:
                self.selfunc_probit   = transform_to_probit(self.selfunc, self.bounds)
                self.log_jacobian_inj = -probit_logJ(self.selfunc, self.bounds, self.probit)
        # MC samples
        self._draw_MC_samples()
        
    def initialise(self):
        """
        Initialise the mixture to initial conditions
        """
        super().initialise()
        self.evaluated_logL = {}
        self._draw_MC_samples()
    
    def _draw_MC_samples(self):
        """
        Draws MC samples for mu and sigma
        """
        self.sigma_MC = np.array([self.invgamma.rvs(self.MC_draws)*(self.exp_sigma[i]**2*(self.a+1)) for i in range(self.dim)]).T
        self.mu_MC    = np.random.uniform(low = self.bounds[:,0], high = self.bounds[:,1], size = (self.MC_draws, self.dim))
        if self.probit:
            self.mu_MC = transform_to_probit(self.mu_MC, self.bounds)
        if self.dim == 1:
            self.sigma_MC = self.sigma_MC.flatten()
            self.mu_MC = self.mu_MC.flatten()
        else:
            rhos = invwishart(df = self.dim+2, scale = np.identity(self.dim)).rvs(size = self.MC_draws)
            rhos = np.array([r/outer_jit(np.sqrt(diag_jit(r)), np.sqrt(diag_jit(r))) for r in rhos])
            self.sigma_MC = np.array([r*outer_jit(s,s) for r, s in zip(rhos, np.sqrt(self.sigma_MC))])
        if self.selfunc is not None:
            # Approximant
            if callable(self.selfunc):
                n_samples = self.MC_draws
                with np.errstate(divide = 'ignore', invalid = 'ignore'):
                    if self.probit:
                        self.log_alpha_factor = np.array([np.log(np.mean(self.selfunc(transform_from_probit(mn(m,s, allow_singular = True).rvs(self.MC_draws), self.bounds)))) for m, s in zip(self.mu_MC, self.sigma_MC)])
                    else:
                        self.log_alpha_factor = np.array([np.log(np.mean(self.selfunc(mn(m,s, allow_singular = True).rvs(self.MC_draws)))) for m, s in zip(self.mu_MC, self.sigma_MC)])
            # Injections
            else:
                n_samples = self.total_inj
                self.log_alpha_factor = np.array([logsumexp_jit(mn(m,s, allow_singular = True).logpdf(self.selfunc_probit) + self.log_jacobian_inj - self.log_inj_pdf) - np.log(self.total_inj) for m, s in zip(self.mu_MC, self.sigma_MC)])
            # Numerical stability: mu+sigma ignored if p_det too small (not enough predicted points to have a reliable value for alpha)
            self.log_alpha_factor[self.log_alpha_factor < np.log(5e-4)] = np.inf
            # Check for NaNs
            self.log_alpha_factor = np.nan_to_num(self.log_alpha_factor, nan = np.inf, posinf = np.inf, neginf = np.inf)
        else:
            self.log_alpha_factor = np.zeros(self.MC_draws)
        
    def add_new_point(self, ev):
        """
        Update the probability density reconstruction adding a new sample
        
        Arguments:
            iterable ev: set of single-event draws from a DPGMM inference
        """
        x = np.random.choice(ev)
        self.stored_pts[len(list(self.stored_pts.keys()))] = x
        self._assign_to_cluster(x)
        self.alpha = _update_alpha(self.alpha, self.n_pts, self.n_cl, self.alpha_0)

    def _cluster_assignment_distribution(self, x, logL_x = None):
        """
        Compute the marginal distribution of cluster assignment for each cluster.
        
        Arguments:
            figaro.mixture.mixture x: sample
            double logL_x:            evaluated log likelihood
        
        Returns:
            dict: p_i for each component
        """
        scores = np.zeros(self.n_cl+1)
        logL_N = np.zeros((self.n_cl+1, self.MC_draws))
        if logL_x is None:
            if self.dim == 1:
                logL_x = evaluate_mixture_MC_draws_1d(self.mu_MC, self.sigma_MC, x.means, x.covs, x.w) - self.log_alpha_factor
            else:
                logL_x = evaluate_mixture_MC_draws(self.mu_MC, self.sigma_MC, x.means, x.covs, x.w) - self.log_alpha_factor
        for i in range(self.n_cl+1):
            if i == 0:
                ss = None
                logL_D = np.zeros(self.MC_draws)
                scores[i] = np.log(self.alpha)
            else:
                ss        = self.mixture[i-1]
                logL_D    = ss.logL_D
                with np.errstate(divide = 'ignore', invalid = 'ignore'):
                    scores[i] = np.log(ss.N)
            if np.isfinite(scores[i]):
                scores[i] += logsumexp_jit(logL_D + logL_x) - logsumexp_jit(logL_D)
            else:
                scores[i] = -np.inf
            logL_N[i]  = logL_D + logL_x
        norm   = logsumexp_jit(scores)
        scores = np.exp(scores-norm)
        return scores, logL_N, logL_x

    def _assign_to_cluster(self, x, pt_id = None, logL_x = None):
        """
        Assign the new sample x to an existing cluster or to a new cluster according to the marginal distribution of cluster assignment.
        
        Arguments:
            np.ndarray x: sample
            int pt_id:    point id
            logL_x:       evaluated log likelihood
        """
        if pt_id is None:
            pt_id = len(list(self.stored_pts.keys()))-1
        self.n_pts += 1
        scores, logL_N, logL_x = self._cluster_assignment_distribution(x, logL_x)
        try:
            cid = np.random.choice(np.arange(-1, self.n_cl), p=scores)
        except ValueError:
            cid = -1
        if cid == -1:
            self.mixture.append(_component_h(x, self.dim, self.prior, logL_N[int(cid)+1], self.mu_MC, self.sigma_MC, self.log_alpha_factor))
            self.N_list.append(1.)
            self.n_cl += 1
            self.assignations[pt_id] = int(self.n_cl)-1
        else:
            self.mixture[int(cid)] = self._add_datapoint_to_component(self.mixture[int(cid)], logL_N[int(cid)+1])
            self.N_list[int(cid)] += 1
            self.assignations[pt_id] = int(cid)
        self.evaluated_logL[pt_id] = logL_x
        # Update weights
        self.w = np.array(self.N_list)
        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            self.w = self.w/self.w.sum()
            self.log_w = np.log(self.w)
        return

    def _remove_from_cluster(self, x, cid, logL_x):
        """
        Remove the sample x from its assigned cluster.
        
        Arguments:
            np.ndarray x:  sample
            int cid:       cluster id
            double logL_x: log likelihood for the point
        """
        self.mixture[int(cid)] = self._remove_datapoint_from_component(self.mixture[int(cid)], self.mixture[int(cid)].logL_D - logL_x)
        self.N_list[int(cid)] -= 1
        self.n_pts            -= 1
        # Update weights
        self.w = np.array(self.N_list)
        with np.errstate(divide = 'ignore', invalid = 'ignore'):
            self.w     = self.w/self.w.sum()
            self.log_w = np.log(self.w)
        return
    
    def _add_datapoint_to_component(self, ss, logL_D):
        """
        Update component parameters after assigning a sample to a component
        
        Arguments:
            component ss: component to update
            double logL_D: log Likelihood denominator
        
        Returns:
            component: updated component
        """
        ss.logL_D = logL_D
        log_norm_D = logsumexp_jit(logL_D)
        
        idx      = np.random.choice(self.MC_draws, p = np.exp(logL_D - log_norm_D))
        ss.mu    = np.copy(self.mu_MC[idx])
        ss.sigma = np.copy(self.sigma_MC[idx])
        if self.dim == 1:
            ss.mu = np.atleast_2d(ss.mu).T
            ss.sigma = np.atleast_2d(ss.sigma).T
        
        ss.N += 1
        ss.log_N_true = np.log(ss.N) - self.log_alpha_factor[idx]
        return ss

    def _remove_datapoint_from_component(self, ss, logL_D):
        """
        Update component parameters after assigning a sample to a component
        
        Arguments:
            component ss: component to update
            double logL_D: log Likelihood denominator
        
        Returns:
            component: updated component
        """
        ss.logL_D = logL_D
        log_norm_D = logsumexp_jit(logL_D)
        
        idx      = np.random.choice(self.MC_draws, p = np.exp(logL_D - log_norm_D))
        ss.mu    = np.copy(self.mu_MC[idx])
        ss.sigma = np.copy(self.sigma_MC[idx])
        if self.dim == 1:
            ss.mu = np.atleast_2d(ss.mu).T
            ss.sigma = np.atleast_2d(ss.sigma).T
        
        ss.N -= 1
        ss.log_N_true = np.log(ss.N) - self.log_alpha_factor[idx]
        return ss

    def build_mixture(self, make_comp = True):
        """
        Instances a mixture class representing the inferred distribution
        
        Arguments:
            bool make_comp:   whether to instantiate the scipy.stats.multivariate_normal components or not
        
        Returns:
            mixture: the inferred distribution
        """
        if self.n_cl == 0:
            raise FIGAROException("You are trying to build an empty mixture - perhaps you called the initialise() method. If you are using the density_from_samples() method, the inferred mixture is returned by that method as an instance of mixture class.")
        idx   = np.where(np.array(self.N_list) > 0)[0]
        N_pts = np.array([comp.log_N_true for comp in np.array(self.mixture)[idx]])
        if self.selfunc is not None:
            log_w, w           = self.log_w, self.w
            self.log_w         = N_pts - logsumexp_jit(N_pts)
            self.w             = np.exp(self.log_w)
            alpha_factor       = self.compute_alpha_factor()
            N_pts              = self.w*self.n_pts/alpha_factor
            self.log_w, self.w = log_w, w
        else:
            alpha_factor = 1.
            N_pts = np.exp(N_pts)
        w = dirichlet(N_pts+self.alpha/self.n_cl).rvs()[0]
        return mixture(np.array([comp.mu for comp in np.array(self.mixture)[idx]]), np.array([comp.sigma for comp in np.array(self.mixture)[idx]]), w, self.bounds, self.dim, (np.array(self.N_list) > 0).sum(), self.n_pts, self.alpha, probit = self.probit, make_comp = make_comp, alpha_factor = alpha_factor)
    
    def compute_alpha_factor(self):
        """
        Compute the integral \int pdet(theta)p(theta|lambda) dtheta conditioned on the specific DPGMM parameters
        
        Returns:
            double: value of the integral
        """
        # Approximant
        if callable(self.selfunc):
            alpha_factor = np.mean(self.selfunc(self.rvs(self.MC_draws)))
        # Injections
        else:
            alpha_factor = np.exp(logsumexp_jit(self.logpdf(self.selfunc) - self.log_inj_pdf) - np.log(self.total_inj))
        return alpha_factor

    def density_from_samples(self, events, make_comp = True):
        """
        Reconstruct the probability density from a set of samples.
        
        Arguments:
            iterable samples: set of single-event draws from DPGMM
            bool make_comp:   whether to instantiate the scipy.stats.multivariate_normal components or not
        
        Returns:
            mixture: the inferred mixture
        """
        np.random.shuffle(events)
        for ev in events:
            self.add_new_point(ev)
        # Random Gibbs walk (if required)
        for id in np.random.choice(self.n_pts, size = self.n_reassignments, replace = True):
            self._reassign_point(id)
        d = self.build_mixture(make_comp = make_comp)
        self.initialise()
        return d

    def _reassign_point(self, id):
        """
        Update the probability density reconstruction reassigining an existing sample
        
        Arguments:
            id: sample id
        """
        x      = self.stored_pts[id]
        cid    = self.assignations[id]
        logL_x = self.evaluated_logL[id]
        self._remove_from_cluster(x, cid, logL_x)
        self._assign_to_cluster(x, id, logL_x)
        self.alpha = _update_alpha(self.alpha, self.n_pts, (np.array(self.N_list) > 0).sum(), self.alpha_0)



import numpy as np
import warnings
warnings.filterwarnings("ignore", "Wswiglal-redir-stdio") # Silence LAL warnings with ipython
from scipy.optimize import newton
from scipy.interpolate import interp1d
from lal._lal import LuminosityDistance, UniformComovingVolumeDensity, ComovingVolumeElement, ComovingVolume, CreateCosmologicalParameters

class CosmologicalParameters:
    """
    Wrapper for LAL functions in a single class.
    
    Arguments:
        double h:   normalised hubble constant h = H0/100 km/Mpc/s
        double om:  matter energy density
        double ol:  cosmological constant density
        double w0:  0th order dark energy equation of state parameter
        double w1:  1st order dark energy equation of state parameter
        double w2:  2nd order dark energy equation of state parameter
        
    Returns:
        CosmologicalParameters: instance of CosmologicalParameters class
    """
    def __init__(self, h, om, ol, w0, w1, w2):
        self.h = h
        self.om = om
        self.ol = ol
        self.w0 = w0
        self.w1 = w1
        self.w2 = w2
        self._CosmologicalParameters = CreateCosmologicalParameters(self.h, self.om, self.ol, self.w0, self.w1, self.w2)
    
    def _vectorise(func):
        def vectorised_func(self, x):
            if hasattr(x, "__iter__"):
                return np.array([func(self, xi) for xi in x])
            else:
                return func(self, x)
        return vectorised_func
    
    @_vectorise
    def LuminosityDistance(self, z):
        return LuminosityDistance(self._CosmologicalParameters, z)
     
    @_vectorise
    def UniformComovingVolumeDensity(self, z):
        return UniformComovingVolumeDensity(z, self._CosmologicalParameters)
    
    @_vectorise
    def ComovingVolumeElement(self, z):
        return ComovingVolumeElement(z, self._CosmologicalParameters)
    
    @_vectorise
    def ComovingVolume(self, z):
        return ComovingVolume(self._CosmologicalParameters, z)
    
    @_vectorise
    def Redshift(self, DL):
        if DL == 0.:
            return 0.
        else:
            def objective(z, self, DL):
                return DL - self.LuminosityDistance(z)
            return newton(objective,1.0,args=(self, DL))

Planck18 = CosmologicalParameters(0.674, 0.315, 0.685, -1, 0, 0)
Planck15 = CosmologicalParameters(0.679, 0.3065, 0.6935, -1, 0, 0)

# Interpolants up to z = 2
z = np.linspace(0,2,1000)
dVdz_approx_planck18 = interp1d(z, Planck18.ComovingVolumeElement(z)/1e9) # In Gpc
dVdz_approx_planck15 = interp1d(z, Planck15.ComovingVolumeElement(z)/1e9) # In Gpc

def _decorator_dVdz(func, approx, z_index, z_max):
    reg_const = (1+z_max)/approx(z_max)
    def decorated_func(x):
        return func(x)*approx(x[:,z_index])/(1+x[:,z_index]) * reg_const
    return decorated_func



import numpy as np
import warnings
import configparser
import optparse

from pathlib import Path
from collections import Counter
from scipy.stats import multivariate_normal as mn

from figaro._numba_functions import *
from figaro.transform import transform_to_probit, transform_from_probit

def recursive_grid(bounds, n_pts, get_1d = False):
    """
    Recursively generates the n-dimensional grid points (extremes are excluded).
    
    Arguments:
        list-of-lists bounds: extremes for each dimension (excluded)
        int n_pts:            number of points for each dimension
        bool get_1d:          return list of 1d-arrays (one per dimension)
        
    Returns:
        np.ndarray: grid
        np.ndarray: differential for each grid
        np.ndarray: list of 1d-arrays (one per dimension)
    """
    bounds = np.atleast_2d(bounds)
    n_pts  = np.atleast_1d(n_pts)
    if len(bounds) == 1:
        d  = np.linspace(bounds[0,0], bounds[0,1], n_pts[0])
        dD = d[1]-d[0]
        if get_1d:
            return np.atleast_2d(d).T, [dD], [d]
        return np.atleast_2d(d).T, [dD]
    else:
        if get_1d:
            grid_nm1, diff, l_1d = recursive_grid(np.array(bounds)[1:], n_pts[1:], get_1d)
        else:
            grid_nm1, diff = recursive_grid(np.array(bounds)[1:], n_pts[1:], get_1d)
        d = np.linspace(bounds[0,0], bounds[0,1], n_pts[0])
        diff.insert(0, d[1]-d[0])
        grid     = []
        for di in d:
            for gi in grid_nm1:
                grid.append([di,*gi])
        if get_1d:
            l_1d.insert(0, d)
            return np.array(grid), diff, l_1d
        return np.array(grid), diff

def rejection_sampler(n_draws, f, bounds, selfunc = None):
    """
    Rejection sampler, allows for a selection function
    
    Arguments:
        int n_draws:      number of draws
        callable f:       probability density to sample from
        iterable bounds:  upper and lower bound
        callable selfunc: selection function, must support numpy arrays
    
    Returns:
        np.ndarray: samples
    """
    n_draws = int(n_draws)
    bounds  = np.atleast_2d(bounds)
    dim     = len(bounds)
    if selfunc is None:
        selfunc = lambda x: 1
    if dim == 1:
        pts = np.linspace(bounds[0,0], bounds[0,1], 1000)
    else:
        pts = np.random.uniform(bounds[:,0], bounds[:,1], size = (3*n_draws,len(bounds)))
    top = np.max(f(pts)*selfunc(pts))
    samples = []
    while len(samples) < n_draws:
        pts   = np.random.uniform(bounds[:,0], bounds[:,1], size = (n_draws,dim))
        if dim == 1:
            pts = pts.flatten()
        probs = f(pts)*selfunc(pts)
        h     = np.random.uniform(0, top, size = n_draws)
        samples.extend(pts[np.where(h < probs)])
    return np.array(samples)[:n_draws]

def get_priors(bounds, samples = None, mean = None, std = None, df = None, k = None, a = None, scale = None, probit = True, hierarchical = False):
    """
    This method takes the prior parameters for the Normal-Inverse-Wishart distribution in the natural space and returns them as parameters in the probit space, ordered as required by FIGARO. In the following, D will denote the dimensionality of the inferred distribution.

    Either four or two parameters are returned:
        * If hierarchical is False, (k, L, df, mu) are returned:
            - df, is the number of degrees of freedom for the Inverse Wishart distribution,. It must be greater than D+1. If this parameter is None or does not satisfy the condition df > D+1, the default value D+2 is used;
            - k is the scale parameter for the multivariate Normal distribution. Suggested values are  k <~ 1e-1. If None, the default value 1e-2 is used.
            - mu is the mean of the multivariate Normal distribution. It can be either estimated from the available samples or passed directly as a 1D array with length D (the keyword argument mean overrides the samples). If None, the default value 0 (corresponding to the parameter space center) is used.
            - L is the expected value for the Inverse Wishart distribution. This parameter can be either:
                - passed as 1D array with shape (D,) or double: vector of standard deviations (if double, it assumes that the same std has to be used for all dimensions) - keyword std;
                - estimated from samples - keyword samples.
        * If hierarchical is True, (L, a) are returned:
            - L is the expected value for the Inverse Gamma distribution. It can be either:
                - passed as 1D array with shape (D,) or double: vector of standard deviations (if double, it assumes that the same std has to be used for all dimensions) - keyword std;
                - estimated from samples - keyword samples;
                - estimated from bounds using the keyword fraction.
            - a is the shape parameter of the gamma distribution. It can be either passed with the dedicated keyword or fixed to the default parameter, 2.

    Arguments:
        np.ndarray bounds:              boundaries for probit transformation
        np.ndarray samples:             2D [DPGMM] or 3D [(H)DPGMM] array with samples
        double or np.ndarray mean:      mean [DPGMM]
        double or np.ndarray std:       expected standard deviation (if double, the same std is used for all dimensions, if np.ndarray must match the number of dimensions) [DPGMM and (H)DPGMM]
        int df:                         degrees of freedom for Inverse Wishart distribution [DPGMM]
        double k:                       scale parameter for Normal distribution [DPGMM]
        double a:                       shape parameter for the Inverse Gamma distribution [(H)DPGMM]
        double scale:                   fraction of samples std [DPGMM]
        bool probit:                    whether the probit transformation will be applied or not
        bool hierarchical:              returns the prior pars for (H)DPGMM rather than for DPGMM
        
    Returns:
        tuple: prior parameters ordered as in (H)/DPGMM
    """
    bounds = np.atleast_2d(bounds)
    dim = len(bounds)
    if scale is None:
        scale = 5.
    if samples is not None:
        if not np.iterable(samples[0]):
            samples = np.atleast_2d(samples).T
        if probit:
            if hierarchical:
                probit_samples = [transform_to_probit(s, bounds) for s in samples]
            else:
                probit_samples = transform_to_probit(samples, bounds)
    if hierarchical:
        if std is not None:
            out_sigma = np.atleast_2d(std)*np.ones((1, dim))
            if probit:
                out_sigma = transform_to_probit(np.mean(bounds, axis = -1)+out_sigma, bounds)
        elif samples is not None:
            if probit:
                all_samples     = np.concatenate(probit_samples, axis = 0)
                events_avg_cov  = np.diag(np.atleast_2d(np.mean([np.cov(ev.T) for ev in probit_samples], axis = 0)))
            else:
                all_samples     = np.concatenate(samples, axis = 0)
                events_avg_cov  = np.diag(np.atleast_2d(np.mean([np.cov(ev.T) for ev in samples], axis = 0)))
            all_samples_cov = np.diag(np.atleast_2d(np.cov(all_samples.T)))
            out_sigma       = (np.sqrt(all_samples_cov - events_avg_cov)/scale).flatten()
        else:
            out_sigma = np.diff(bounds, axis = -1)/scale
            if probit:
                out_sigma = transform_to_probit(np.mean(bounds, axis = -1)+out_sigma, bounds)
        out_sigma = out_sigma.flatten()
        if a is not None:
            out_a = a
        else:
            out_a = 2.
        return (out_sigma, out_a)
    else:
        # DF
        if df is not None and df > dim+2:
            df_out = df
        else:
            df_out = dim+2
        # Mu
        if mean is not None:
            mean = np.atleast_1d(mean)
            if not np.prod(bounds[:,0] < mean) & np.prod(mean < bounds[:,1]):
                raise ValueError("Mean is outside of the given bounds")
            if probit:
                mu_out = transform_to_probit(mean, bounds)
            else:
                mu_out = mean
        elif samples is not None:
            if probit:
                mu_out = np.atleast_1d(np.mean(probit_samples, axis = 0))
            else:
                mu_out = np.atleast_1d(np.mean(samples, axis = 0))
        else:
            if probit:
                mu_out = np.zeros(dim)
            else:
                mu_out = np.atleast_1d(np.mean(bounds, axis = 1))
        # L
        if std is not None:
            L_out = np.identity(dim)*std**2
        elif samples is not None:
            if probit:
                cov_samples = np.atleast_2d(np.cov(probit_samples.T))
            else:
                cov_samples = np.atleast_2d(np.cov(samples.T))
            L_out = rescale_matrix(cov_samples, scale**2)
        else:
            if probit:
                sigma = transform_to_probit(np.atleast_2d(np.mean(bounds, axis = -1)+np.diff(bounds, axis = -1).flatten()/scale), bounds)[0]
                L_out = np.identity(dim)*sigma**2
            else:
                L_out = np.identity(dim)*(np.diff(bounds, axis = -1).flatten()/scale)**2
        # k
        if k is not None:
            k_out = k
        else:
            if samples is not None:
                if probit:
                    cov_samples = np.atleast_2d(np.cov(probit_samples.T))
                else:
                    cov_samples = np.atleast_2d(np.cov(samples.T))
                k_out = np.min(np.diag(L_out)/np.diag(cov_samples))
            else:
                k_out = 1./(scale)

        return (k_out, L_out, df_out, mu_out)

def gradient_median(x, draws):
    """
    Computes the gradient of the median distribution (slow but numerically stable)
    
    Arguments:
        np.ndarray x:   points to evaluate the median at
        iterable draws: container for mixture instances
    
    Returns:
        np.ndarray: gradient
    """
    return np.median([d.gradient(x) for d in draws], axis = 0)

def rvs_median(draws, size = 1):
    """
    Generates samples from median distribution of a set of draws.
    
    Arguments:
        iterable draws: container for mixture instances
        int size:       number of samples
    
    Returns:
        np.ndarray: samples
    """
    idx = np.random.choice(np.arange(len(draws)), size = int(size))
    ctr = Counter(idx)
    samples = np.empty(shape = (1, draws[0].dim))
    for i, n in zip(ctr.keys(), ctr.values()):
        samples = np.concatenate((samples, draws[i].rvs(n)))
    return samples[1:]
    
def make_gaussian_mixture(mu, cov, bounds, out_folder = '.', names = None, save = False, save_samples = False, n_samps = 3000, probit = True, ext = 'json'):
    """
    Builds mixtures composed of equally-weighted Gaussian distribution.
    WARNING: due to the probit coordinate change, a Gaussian distribution in the natural space does not correspond to a Gaussian distribution in the probit space.
    The resulting distributions in probit space, therefore, are just an approximation. This approximation holds for distributions which are far from boundaries.
    In general, a more robust (but slower) approach would be to draw samples from each original Gaussian distribution and to use them to make a hierarchical inference.
    
    Arguments:
        np.ndarray mu:     mean for each Gaussian distribution
        np.ndarray cov:    covariance matrix for each Gaussian distribution
        np.ndarray bounds: boundaries for probit transformation
        str out_folder:    output folder
        bool save:         whether to save the draws or not
        bool save_samples: whether to save the samples or not
        int n_samps:       number of samples to estimate mean and covariance in probit space
        bool probit:       whether to use the probit transformation or not
        str ext:           file extension (pkl or json)
    
    Returns:
        np.ndarray: mixtures
    """
    # Here to avoid circular import
    from figaro.mixture import mixture
    from figaro.load import save_density
    bounds = np.atleast_2d(bounds)
    dim    = len(bounds)
    
    out_folder = Path(out_folder)
    if not out_folder.exists():
        out_folder.mkdir()
    
    if save:
        draws_folder = Path(out_folder, 'draws')
        if not draws_folder.exists():
            draws_folder.mkdir()
    
    if save_samples:
        events_folder = Path(out_folder, 'events')
        if not events_folder.exists():
            events_folder.mkdir()
    
    mixtures = []
    for i, (means, covs) in enumerate(zip(mu, cov)):
        means = np.atleast_2d(means).reshape(-1,dim)
        covs  = np.atleast_3d(covs).reshape(-1,dim,dim)
        if save_samples:
            samples = np.empty(shape = (1,len(bounds)))
        mm = []
        cc = []
        for m, c in zip(means, covs):
            if probit:
                ss = np.atleast_2d(mn(m, c, allow_singular = True).rvs(n_samps))
                # 1D issue
                if c.shape == (1,) or c.shape == (1,1):
                    ss = ss.T
                # Keeping only samples within bounds
                ss = ss[np.where((np.prod(bounds[:,0] < ss, axis = 1) & np.prod(ss < bounds[:,1], axis = 1)))]
                if save_samples:
                    samples = np.concatenate((samples, ss))
                # Probit samples
                p_ss = transform_to_probit(ss, bounds)
                mm.append(np.mean(p_ss, axis = 0))
                cc.append(np.atleast_2d(np.cov(p_ss.T)))
            else:
                mm.append(m)
                cc.append(c)
                if save:
                    ss = np.atleast_2d(mn(m, c, allow_singular = True).rvs(n_samps))
                    if ss.shape[0] == 1:
                        ss = ss.T
                    samples = np.concatenate((samples, ss))
        if save_samples:
            if names is not None:
                name = names[i]
            else:
                name = 'event_{0}'.format(i+1)
            np.savetxt(Path(events_folder, name+'.txt'), samples[1:])
        mix = mixture(np.atleast_2d(mm), np.atleast_3d(cc), np.ones(len(means))/len(means), bounds, len(bounds), len(means), 0, probit = probit, alpha = 1.)
        if save:
            save_density([mix], draws_folder, name, ext)
        mixtures.append([mix])
    mixtures = np.array(mixtures)
    
    if save:
        # Circular import
        from figaro.load import save_density
        save_density(mixtures, draws_folder, 'posteriors_single_event', ext)
    
    return mixtures

#-------------#
#   Options   #
#-------------#

def save_options(options, out_folder, name = None):
    """
    Saves options for the run (reproducibility)
    
    Arguments:
        optparser.Options options: options
        str or Path out_folder:    folder where to save the option file
        str name:                  name of the run
    """
    if name is None:
        filename = 'options.ini'
    else:
        filename = 'options_{0}.ini'.format(name)
    dd = {key:str(val) for (key, val) in vars(options).items() if not key == 'config'}
    config = configparser.ConfigParser()
    config.read_dict({'DEFAULT':dd})
    with open(filename, 'w') as f:
        config.write(f)

def load_options(opts, parser):
    """
    Loads options for the run (reproducibility)
    
    Arguments:
        optparser.Options opts:       options object
        optparse.OptionParser parser: parser object
    
    Returns:
        optparser.Options: options
    """
    config = configparser.ConfigParser()
    config.read(opts.config)
    defaults = {}
    # Convert None and bools appropriately
    for (key, val) in config.defaults().items():
        if val == 'None':
            val = None
        elif val == 'True':
            val = True
        elif val == 'False':
            val = False
        defaults[key] = val
    parser.set_defaults(**defaults)
    opts, _ = parser.parse_args()
    return opts



import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
from collections import Counter
from scipy.stats import poisson, norm
from figaro.cosmology import dVdz_approx_planck18

def sample_rate(draws, n_obs, selfunc, T, volume = None, size = None, each = False, n_draws = 1e4, dvdz = None, z_index = -1, normalise_alpha = False):
    """
    Compute the integrated rate given a set of draws. If volume keyword is not used, returns the expected number of events before filtering.
    
    Arguments:
        iterable draws:       container for mixture instances
        int n_obs:            number of observed events
        callable selfunc:     selection function approximant
        double T:             duration of the observations
        double volume:        surveyed volume. If None, it is estimated.
        int size:             number of samples
        bool each:            sample one point from each draw
        int n_draws:          number of MC draws
        callable dvdz         function to compute the comoving volume element (in Gpc^3)
        int z_index           redshift parameter index (default: last dimension)
        bool normalise_alpha: re-normalise alpha (required if the selection function includes dV/dz*(1+z)^-1)
    
    Returns:
        np.ndarray: integrated rate samples
    """
    if size is None or each:
        size = len(draws)
    if volume is None:
        volume = sample_VT(draws, selfunc, T, size = size, each = each, dvdz = dvdz, z_index = z_index)
    if normalise_alpha:
        normalise_alpha_factor(draws, dvdz = dvdz, z_index = z_index, n_draws = n_draws)
    N_exp   = [poisson(int(n_obs/d.alpha_factor)) for d in draws]
    if each:
        return np.array([N.mean()/v for N, v in zip(N_exp, volume)])
    idx     = np.random.choice(np.arange(len(N_exp)), size = int(size))
    ctr     = Counter(idx)
    samples = np.empty(shape = (1,))
    for i, n in zip(ctr.keys(), ctr.values()):
        samples = np.concatenate((samples, N_exp[i].rvs(n)))
    return np.divide(samples[1:], volume)

def normalise_alpha_factor(draws, dvdz = None, z_index = -1, z_max = None, n_draws = 1e4):
    """
    Normalise the alpha factor for each DPGMM realisation. This is required if during the reconstruction the selection function included the dV/dz*(1+z)^-1 term.
    
    Arguments:
        iterable draws: container for mixture instances
        callable dvdz:  function to compute the comoving volume element (in Gpc^3)
        int z_index:    redshift parameter index (default: last dimension)
        double z_max:   max redshift value
        int n_draws:    number of MC draws
    """
    if dvdz is None:
        dvdz = dVdz_approx_planck18
    if z_max is not None:
        reg_const = (1+z_max)/dvdz(z_max)
    else:
        reg_const = 1.
    for d in draws:
        ss = d.rvs(int(n_draws))
        d.alpha_factor /= np.mean(reg_const*dvdz(ss[:,z_index])/(1+ss[:,z_index]))

def sample_VT(draws, selfunc, T, size = None, n_draws = 1e4, each = False, dvdz = None, z_index = -1):
    """
    Draw samples from the VT distribution given the astrophysical distribution of quantities, the selection function and the observing time.
    See Eqs. (21) of Kapadia et al (2020) (https://arxiv.org/pdf/1903.06881.pdf).
    The calibration uncertainty is estimated as in Sec. 5 - Eq. (17) - of Abbott et al (2016) (https://arxiv.org/pdf/1606.03939.pdf)
    
    Arguments:
        iterable draws:   container for mixture instances
        callable selfunc: selection function approximant
        double T:         duration of the observations
        int size:         number of samples
        int n_draws:      number of MC draws
        bool each:        sample one point from each draw
        callable dvdz     function to compute the comoving volume element (in Gpc^3)
        int z_index       redshift parameter index (default: last dimension)
    
    Returns:
        np.ndarray: VT samples
    """
    draws = np.atleast_1d(draws)
    if size is None or each:
        size = len(draws)
    if dvdz is None:
        dvdz = dVdz_approx_planck18
    VT_dist = []
    for d in draws:
        ss     = d.rvs(int(n_draws))
        comvol = dvdz(ss[:,z_index])
        sf     = selfunc(ss)
        VT_s   = T*sf*comvol/(1+ss[:,z_index])
        VT     = np.mean(VT_s)
        S      = np.sqrt(np.cov(VT_s)/n_draws)
        VT_dist.append(norm(np.log(VT),S/VT))
    if each:
        return np.array([np.exp(VT_i.mean()) for VT_i in VT_dist])
    else:
        idx     = np.random.choice(np.arange(len(draws)), size = int(size))
        ctr     = Counter(idx)
        samples = np.empty(shape = (1,))
        for i, n in zip(ctr.keys(), ctr.values()):
            samples = np.concatenate((samples, np.exp(VT_dist[i].rvs(n))))
        return samples[1:]

def plot_integrated_rate(samples, out_folder = '.', name = 'density', volume_unit = '\\mathrm{Gpc}^{-3}\\mathrm{yr}^{-1}', show = False, save = True, true_value = None, true_label = None):
    """
    Plot the integrated rate distribution.
    
    Arguments:
        np.ndarray samples:     rate samples
        str or Path out_folder: output folder
        str name:               name to be given to output
        str volume_unit:        LaTeX-style volume unit, for plotting purposes
        bool save:              whether to save the plot or not
        bool show:              whether to show the plot during the run or not
        iterable true_value:    true value to plot
        iterable levels:        credible levels to plot
        str true_label:         label to assign to the reconstruction
    
    Returns:
        matplotlib.figure.Figure: figure with the plot
    """
    if true_label is None:
        true_label = '$\\mathcal{R}_\\mathrm{true}$'
    else:
        true_label = f'${true_label}$'
    if volume_unit is not None:
        xlabel = '$\\mathcal{R}\ ['+volume_unit+']$'
    else:
        xlabel = '$\\mathcal{R}$'
    fig, ax = plt.subplots()
    if true_value is not None:
        ax.axvline(true_value, color = 'firebrick', dashes = (5,5), label = true_label)
    ax.hist(samples, histtype = 'step', density = True)
    if true_value is not None:
        ax.legend(loc = 0)
    ax.set_xlabel(xlabel)
    ax.set_ylabel('$p(\\mathcal{R})$')
    if save:
        fig.savefig(Path(out_folder, 'integrated_rate_{0}.pdf'.format(name)), bbox_inches = 'tight')
    if show:
        plt.show()
    plt.close()
    return fig

def plot_differential_rate(draws, rate, injected = None, bounds = None, out_folder = '.', name = 'density', n_pts = 1000, label = None, unit = None, volume_unit = '\\mathrm{Gpc}^{-3}\\mathrm{yr}^{-1}', hierarchical = False, show = False, save = True, true_value = None, true_value_label = '\mathrm{True\ value}', injected_label = '\mathrm{Simulated}', median_label = None, fig = None, colors = ['steelblue', 'darkturquoise', 'mediumturquoise']):
    """
    Plot the recovered 1D distribution along with the injected distribution and samples from the true distribution (both if available).
    
    Arguments:
        iterable draws:                  container for mixture instances
        np.ndarray rate:                 rate samples or value
        callable or np.ndarray injected: injected distribution (if available)
        iterable bounds:                 bounds for the recovered distribution. If None, bounds from mixture instances are used.
        str or Path out_folder:          output folder
        str name:                        name to be given to outputs
        int n_pts:                       number of points for linspace
        str label:                       LaTeX-style quantity label, for plotting purposes
        str unit:                        LaTeX-style quantity unit, for plotting purposes
        str volume_unit:                 LaTeX-style volume unit, for plotting purposes
        bool hierarchical:               hierarchical inference, for plotting purposes
        bool save:                       whether to save the plots or not
        bool show:                       whether to show the plots during the run or not
        float true_value:                true value to infer
        str true_value_label:            label to assign to the true value marker
        str injected_label:              label to assign to the injected distribution
        str median_label:                label to assign to the reconstruction
        matplotlib.figure.Figure fig:    figure to use for plotting. Must have (dim,dim) axes.
        list-of-str colors:              list of colors for median, 68% and 90% credible regions
    
    Returns:
        matplotlib.figure.Figure: figure with the plot
    """
    if median_label is None:
        if hierarchical:
            median_label = '\mathrm{(H)DPGMM}'
        else:
            median_label = '\mathrm{DPGMM}'
    
    if not hasattr(rate, '__iter__'):
        rate = np.ones(len(draws))*rate
    
    color_med, color_68, color_90 = colors
    
    all_bounds = np.atleast_2d([d.bounds[0] for d in draws])
    x_min = np.max(all_bounds[:,0])
    x_max = np.min(all_bounds[:,1])
    
    probit = np.array([d.probit for d in draws]).any()
    
    if bounds is not None:
        if not bounds[0] >= x_min and probit:
            warnings.warn("The provided lower bound is invalid for at least one draw. {0} will be used instead.".format(x_min))
        else:
            x_min = bounds[0]
        if not bounds[1] <= x_max and probit:
            warnings.warn("The provided upper bound is invalid for at least one draw. {0} will be used instead.".format(x_max))
        else:
            x_max = bounds[1]

    if fig is None:
        fig, ax = plt.subplots()
    else:
        ax = fig.axes[0]

    xlim  = (x_min, x_max)
    x     = np.linspace(x_min, x_max, n_pts)
    dx    = x[1]-x[0]
    probs = np.array([r*d.pdf(x) for d, r in zip(draws, rate)])
    percentiles = [50, 5, 16, 84, 95]
    p = {}
    for perc in percentiles:
        p[perc] = np.percentile(probs, perc, axis = 0)
    
    ax.set_yscale('log')
    # CR
    ax.fill_between(x, p[95], p[5], color = color_90, alpha = 0.25)
    ax.fill_between(x, p[84], p[16], color = color_68, alpha = 0.25)
    # Injection (if available)
    if injected is not None:
        if callable(injected):
            p_x = injected(x)
        else:
            p_x = injected
        if injected_label is not None:
            injected_label = '$'+injected_label+'$'
        ax.plot(x, p_x, lw = 0.5, color = 'red', label = injected_label)
        
    # Median
    if true_value is not None:
        if true_value_label is not None:
            true_value_label = '$'+true_value_label+'$'
        ax.axvline(true_value, ls = '--', color = 'firebrick', dashes = (5,5), lw = 0.5, label = true_value_label)
    ax.plot(x, p[50], lw = 0.7, color = color_med, label = '${0}$'.format(median_label))
    if label is None:
        label = 'x'
    if unit is None or unit == '':
        ax.set_xlabel('${0}$'.format(label))
        unit_label = volume_unit
    else:
        ax.set_xlabel('${0}\ [{1}]$'.format(label, unit))
        unit_label = volume_unit +unit+'^{-1}'
    y_label = '$\\frac{\\mathrm{d}\\mathcal{R}}{\\mathrm{d}'+label+'}\ ['+unit_label+']$'
    ax.set_ylabel(y_label)
    ax.set_ylim(bottom = 1e-5, top = np.max(p[95])*1.1)
    ax.legend(loc = 0)
    
    fig.align_labels()
    
    if save:
        plot_folder = out_folder
        log_folder  = out_folder
        txt_folder  = out_folder
        fig.savefig(Path(log_folder, 'log_differential_rate_{0}.pdf'.format(name)), bbox_inches = 'tight')
        ax.set_yscale('linear')
        ax.autoscale(True)
        fig.savefig(Path(plot_folder, 'differential_rate_{0}.pdf'.format(name)), bbox_inches = 'tight')
        np.savetxt(Path(txt_folder, 'prob_differential_rate_{0}.txt'.format(name)), np.array([x, p[50], p[5], p[16], p[84], p[95]]).T, header = 'x 50 5 16 84 95')
    if show:
        ax.set_yscale('linear')
        ax.autoscale(True)
        plt.show()
    plt.close()
    return fig

